import { NetworkDelay, RtdbOrder, isAdminSdk, SDK, isFirestoreDatabase, ApiKind, isMockConfig } from '@forest-fire/types';
import { uuid, between, atRandom, get } from 'native-dash';
import merge from 'deepmerge';
import { createError } from 'brilliant-errors';
import { validate } from 'email-validator';
import { deepEqual } from 'fast-equals';
import copy from 'fast-copy';
import { key } from 'firebase-key';
import { SerializedRealTimeQuery, slashNotation } from '@forest-fire/serialized-query';
import { hashToArray, arrayToHash } from 'typed-conversions';

var SortOrder;
(function (SortOrder) {
    SortOrder[SortOrder["asc"] = 0] = "asc";
    SortOrder[SortOrder["desc"] = 1] = "desc";
})(SortOrder || (SortOrder = {}));

class FireMockError extends Error {
    constructor(message, classification = 'firemock/error') {
        super(message);
        this.firemodel = true;
        const [cat, subcat] = classification.split('/');
        this.code = subcat || 'error';
        this.name = classification;
    }
}

function isUserCredential(u) {
    if (typeof u !== 'object')
        return false;
    const check = u;
    return check.user && check.credential && check.additionalUserInfo
        ? true
        : false;
}
/**
 * Boolean flag and type guard that checks whether this is Firemock `IMock` user
 * definition.
 */
function isMockUserRecord(u) {
    if (typeof u !== 'object')
        return false;
    const check = u;
    return check.kind && check.kind === 'MockUserRecord';
}
/**
 * Boolean flag and type guard that checks whether this is Admin SDK's `UserRecord`.
 *
 * Note: the `IMockUser` is a superset of this.
 */
function isUserRecord(u) {
    const check = u;
    return (check.metadata &&
        check.providerData &&
        check.emailVerified !== undefined &&
        check.kind === undefined);
}
/**
 * Boolean flat to detect if user-like object is a `User` from the client SDK
 */
function isUser(user) {
    return !isMockUserRecord(user) &&
        !isUserRecord(user) &&
        user.uid !== undefined
        ? true
        : false;
}

/**
 * takes a partial user auth and adds enough to make it officially
 * a full UserCrediental
 */
function completeUserCredential(partial) {
    const fakeUserCredential = {
        user: {
            // TODO: Implement
            // ...clientApiUser,
            displayName: '',
            email: '',
            isAnonymous: true,
            metadata: {},
            phoneNumber: '',
            photoURL: '',
            providerData: [],
            providerId: '',
            refreshToken: '',
            // TODO: Check if this util method work as expected
            uid: uuid()
        },
        additionalUserInfo: {
            isNewUser: false,
            profile: '',
            providerId: '',
            username: 'fake',
        },
        operationType: '',
        credential: {
            signInMethod: 'fake',
            providerId: 'fake',
            toJSON: () => '', // added recently
        },
        // TODO: Implement
    };
    return merge(fakeUserCredential, partial);
}
const fakeApplicationVerifier = {
    // eslint-disable-next-line @typescript-eslint/require-await
    async confirm(verificationCode) {
        return completeUserCredential({});
    },
    verificationId: 'verification',
};

function createAnonymousUser(uid) {
    return {
        uid,
        isAnonymous: true,
        // TODO: Implement 
        // ...authUsersApi,
    };
}

/**
 * Converts a known user to the `User` format used in the Firebase client SDK
 */
function toUser(mockUser) {
    // TODO: implement
    const user = {
        uid: mockUser.uid,
        email: mockUser.email,
        emailVerified: mockUser.emailVerified,
        displayName: mockUser.displayName,
        isAnonymous: mockUser.isAnonymous,
        providerId: mockUser.providerId,
        providerData: mockUser.providerData,
        // TODO: implement
    };
    return {
        ...user,
        toJSON: () => ({
            uid: user.uid,
            isAnonymous: user.isAnonymous,
            email: user.email,
            emailVerified: user.emailVerified,
            phoneNumber: user.phoneNumber,
            photoURL: user.photoURL,
        }),
    };
}

function toUserCredential(user) {
    return {
        credential: {
            providerId: user.providerId,
            signInMethod: '',
            toJSON() {
                return { providerId: user.providerId };
            },
        },
        /** the client SDK's `User` type */
        user: toUser(user),
        additionalUserInfo: {
            isNewUser: false,
            profile: {},
            providerId: user.providerId,
            username: user.email,
        },
        /** not sure what this is intended to be */
        operationType: '',
    };
}

function toUserRecord(user) {
    if (!user) {
        return undefined;
    }
    const u = { ...user };
    delete u.password;
    delete u.providerId;
    return u;
}

const dotify = (path) => {
    const dotPath = path.replace(/[\\/]/g, '.');
    return removeDotsAtExtremes(dotPath.slice(0, 1) === '.' ? dotPath.slice(1) : dotPath);
};
function dotifyKeys(obj) {
    const result = {};
    Object.keys(obj).forEach((key) => {
        result[dotify(key)] = obj[key];
    });
    return result;
}
function removeDotsAtExtremes(path) {
    const front = path.slice(0, 1) === '.' ? path.slice(1) : path;
    return front.slice(-1) === '.' ? front.slice(0, front.length - 1) : front;
}

/**
 * provides a friendly way to pause execution when using
 * async/await symantics
 */
async function wait(ms) {
    return new Promise((resolve) => setTimeout(resolve, ms));
}

/**
 * Provides a string lookup of a AWS region's
 * geographic name based on it's more technical AWS region
 * name.
 *
 * For instance, `us-east-1` looks up to "N. Virginia", etc.
 */
var AwsRegionName;
(function (AwsRegionName) {
    // US
    AwsRegionName["us-east-1"] = "N. Virginia";
    AwsRegionName["us-east-2"] = "Ohio";
    AwsRegionName["us-west-1"] = "N. California";
    AwsRegionName["us-west-2"] = "Oregon";
    // Africa
    AwsRegionName["af-south-1"] = "Cape Town";
    // AP
    AwsRegionName["ap-east-1"] = "Hong Kong";
    AwsRegionName["ap-south-1"] = "Mumbai";
    AwsRegionName["ap-northeast-1"] = "Tokyo";
    AwsRegionName["ap-northeast-2"] = "Seoul";
    AwsRegionName["ap-northeast-3"] = "Osaka";
    AwsRegionName["ap-southeast-1"] = "Singapore";
    AwsRegionName["ap-southeast-2"] = "Sydney";
    // Canada
    AwsRegionName["ca-central-1"] = "Canada (central)";
    // EU
    AwsRegionName["eu-central-1"] = "Frankfurt";
    AwsRegionName["eu-west-1"] = "Ireland";
    AwsRegionName["eu-west-2"] = "London";
    AwsRegionName["eu-west-3"] = "Paris";
    AwsRegionName["eu-north-1"] = "Stockholm";
    AwsRegionName["eu-south-1"] = "Milan";
    // Middle East
    AwsRegionName["me-south-1"] = "Bahrain";
    // South America
    AwsRegionName["sa-east-1"] = "S\u0101o Paulo";
})(AwsRegionName || (AwsRegionName = {}));

var HttpStatusCodes;
(function (HttpStatusCodes) {
    /**
     * The client SHOULD continue with its request. This interim response is used to inform
     * the client that the initial part of the request has been received and has not yet
     * been rejected by the server. The client SHOULD continue by sending the remainder
     * of the request or, if the request has already been completed, ignore this response.
     * The server MUST send a final response after the request has been completed.
     */
    HttpStatusCodes[HttpStatusCodes["Continue"] = 100] = "Continue";
    /** The request has succeeded. */
    HttpStatusCodes[HttpStatusCodes["Success"] = 200] = "Success";
    /**
     * The request has been fulfilled and resulted in a new resource being created. The newly
     * created resource can be referenced by the URI(s) returned in the entity of the response,
     * with the most specific URI for the resource given by a Location header field. The response
     * SHOULD include an entity containing a list of resource characteristics and location(s) from
     * which the user or user agent can choose the one most appropriate. The entity format is
     * specified by the media type given in the Content-Type header field. The origin server MUST
     * create the resource before returning the `201` status code. If the action cannot be carried
     * out immediately, the server SHOULD respond with `202` (Accepted) response instead.
     *
     * A `201` response MAY contain an ETag response header field indicating the current value of
     * the entity tag for the requested variant just created.
  
     */
    HttpStatusCodes[HttpStatusCodes["Created"] = 201] = "Created";
    /**
     * The request has been accepted for processing, but the processing has not been completed.
     * The request might or might not eventually be acted upon, as it might be disallowed when
     * processing actually takes place. There is no facility for re-sending a status code from an
     * asynchronous operation such as this.
     *
     * The 202 response is intentionally non-committal. Its purpose is to allow a server to accept
     * a request for some other process (perhaps a batch-oriented process that is only run once
     * per day) without requiring that the user agent's connection to the server persist until the
     * process is completed. The entity returned with this response SHOULD include an indication
     * of the request's current status and either a pointer to a status monitor or some estimate
     * of when the user can expect the request to be fulfilled.
     */
    HttpStatusCodes[HttpStatusCodes["Accepted"] = 202] = "Accepted";
    /**
     * The server has fulfilled the request but does not need to return an entity-body, and might
     * want to return updated meta-information. The response MAY include new or updated
     * meta-information in the form of entity-headers, which if present SHOULD be associated with
     * the requested variant.
     *
     * If the client is a _user agent_, it **SHOULD NOT** change its document view from that which
     * caused the request to be sent. This response is primarily intended to allow input for
     * actions to take place without causing a change to the user agent's active document view,
     * although any new or updated metainformation **SHOULD** be applied to the document
     * currently in the user agent's active view.
     *
     * The `204` response **MUST NOT** include a `message-body`, and thus is always terminated
     * by the first empty line after the header fields.
     */
    HttpStatusCodes[HttpStatusCodes["NoContent"] = 204] = "NoContent";
    HttpStatusCodes[HttpStatusCodes["MovedPermenantly"] = 301] = "MovedPermenantly";
    HttpStatusCodes[HttpStatusCodes["TemporaryRedirect"] = 307] = "TemporaryRedirect";
    /**
     * If the client has performed a conditional GET request and access is allowed, but the
     * document has not been modified, the server SHOULD respond with this status code. The
     * `304` response MUST NOT contain a _message-body_, and thus is always terminated by the
     * first empty line after the header fields.
     */
    HttpStatusCodes[HttpStatusCodes["NotModified"] = 304] = "NotModified";
    /**
     * The request could not be understood by the server due to malformed syntax.
     * The client SHOULD NOT repeat the request without modifications.
     */
    HttpStatusCodes[HttpStatusCodes["BadRequest"] = 400] = "BadRequest";
    /**
     * The request requires user authentication. The response MUST include a WWW-Authenticate
     * header field containing a challenge applicable to the requested resource.
     * The client MAY repeat the request with a suitable Authorization header field. If the
     * request already included Authorization credentials, then the `401`
     * response indicates that authorization has been refused for those credentials. If the `401`
     * response contains the same challenge as the prior response, and the user agent has already
     * attempted authentication at least once, then the user SHOULD be presented the entity that
     * was given in the response, since that entity might include relevant diagnostic information.
     */
    HttpStatusCodes[HttpStatusCodes["Unauthorized"] = 401] = "Unauthorized";
    HttpStatusCodes[HttpStatusCodes["PaymentRequired"] = 402] = "PaymentRequired";
    /**
     * The request could not be understood by the server due to malformed syntax. The client
     * SHOULD NOT repeat the request without modifications.
     */
    HttpStatusCodes[HttpStatusCodes["Forbidden"] = 403] = "Forbidden";
    /**
     * The server has not found anything matching the Request-URI. No indication is given of
     * whether the condition is temporary or permanent. The `410` (Gone) status code SHOULD be
     * used if the server knows, through some internally configurable mechanism, that an old
     * resource is permanently unavailable and has no forwarding address.
     *
     * This status code is commonly used when the server does not wish to reveal exactly
     * why the request has been refused, or when no other response is applicable.
     */
    HttpStatusCodes[HttpStatusCodes["NotFound"] = 404] = "NotFound";
    /**
     * The method specified in the Request-Line is not allowed for the resource identified
     * by the Request-URI. The response MUST include an Allow header containing a list of
     * valid methods for the requested resource.
     */
    HttpStatusCodes[HttpStatusCodes["MethodNotAllowed"] = 405] = "MethodNotAllowed";
    /**
     * The client did not produce a request within the time that the server was
     * prepared to wait. The client MAY repeat the request without modifications
     * at any later time.
     */
    HttpStatusCodes[HttpStatusCodes["RequestTimeout"] = 408] = "RequestTimeout";
    /**
     * The request could not be completed due to a conflict with the current state of the
     * resource. This code is only allowed in situations where it is expected that the
     * user might be able to resolve the conflict and resubmit the request. The response
     * body SHOULD include enough information for the user to recognize the source of the
     * conflict. Ideally, the response entity would include enough information for the
     * user or user agent to fix the problem; however, that might not be possible and
     * is not required.
     *
     * Conflicts are most likely to occur in response to a PUT request. For example,
     * if versioning were being used and the entity being PUT included changes to a resource
     * which conflict with those made by an earlier (third-party) request, the server might
     * use the 409 response to indicate that it can't complete the request. In this case,
     * the response entity would likely contain a list of the differences between the
     * two versions in a format defined by the response Content-Type.
     */
    HttpStatusCodes[HttpStatusCodes["Conflict"] = 409] = "Conflict";
    /**
     * The requested resource is no longer available at the server and no forwarding address
     * is known. This condition is expected to be considered permanent. Clients with link
     * editing capabilities SHOULD delete references to the Request-URI after user approval.
     * If the server does not know, or has no facility to determine, whether or not the
     * condition is permanent, the status code 404 (Not Found) SHOULD be used instead.
     * This response is cacheable unless indicated otherwise.
     *
     * The 410 response is primarily intended to assist the task of web maintenance by
     * notifying the recipient that the resource is intentionally unavailable and that
     * the server owners desire that remote links to that resource be removed. Such an
     * event is common for limited-time, promotional services and for resources belonging
     * to individuals no longer working at the server's site. It is not necessary to mark
     * all permanently unavailable resources as "gone" or to keep the mark for any length
     * of time -- that is left to the discretion of the server owner.
     */
    HttpStatusCodes[HttpStatusCodes["Gone"] = 410] = "Gone";
    /**
     * Indicates that the server refuses to brew coffee because it is, permanently, a teapot.
     * A combined coffee/tea pot that is temporarily out of coffee should instead return 503.
     * This error is a reference to Hyper Text Coffee Pot Control Protocol defined in April
     * Fools' jokes in 1998 and 2014.
     */
    HttpStatusCodes[HttpStatusCodes["IAmATeapot"] = 418] = "IAmATeapot";
    HttpStatusCodes[HttpStatusCodes["UnprocessableEntity"] = 422] = "UnprocessableEntity";
    /**
     * The 429 status code indicates that the user has sent too many requests in a given
     * amount of time ("rate limiting").
     */
    HttpStatusCodes[HttpStatusCodes["TooManyRequests"] = 429] = "TooManyRequests";
    /**
     * The server encountered an unexpected condition which prevented it from fulfilling
     * the request.
     */
    HttpStatusCodes[HttpStatusCodes["InternalServerError"] = 500] = "InternalServerError";
    /**
     * The server does not support the functionality required to fulfill the request. This
     * is the appropriate response when the server does not recognize the request method
     * and is not capable of supporting it for any resource.
     */
    HttpStatusCodes[HttpStatusCodes["NotImplemented"] = 501] = "NotImplemented";
    /**
     * The server, while acting as a gateway or proxy, received an invalid response from
     * the upstream server it accessed in attempting to fulfill the request.
     */
    HttpStatusCodes[HttpStatusCodes["BadGateway"] = 502] = "BadGateway";
    /**
     * Indicates that the server is not ready to handle the request.
     *
     * Common causes are a server that is down for maintenance or that is overloaded.
     * This response should be used for temporary conditions and the `Retry-After` HTTP
     * header should, if possible, contain the estimated time for the recovery of the
     * service.
     */
    HttpStatusCodes[HttpStatusCodes["ServiceUnavailable"] = 503] = "ServiceUnavailable";
    HttpStatusCodes[HttpStatusCodes["GatewayTimeout"] = 504] = "GatewayTimeout";
    /**
     * The 511 status code indicates that the client needs to authenticate to gain
     * network access.
     *
     * The response representation SHOULD contain a link to a resource that allows
     * the user to submit credentials (e.g. with a HTML form).
     *
     * Note that the 511 response SHOULD NOT contain a challenge or the login interface
     * itself, because browsers would show the login interface as being associated with
     * the originally requested URL, which may cause confusion.
     */
    HttpStatusCodes[HttpStatusCodes["AuthenticationRequired"] = 511] = "AuthenticationRequired";
})(HttpStatusCodes || (HttpStatusCodes = {}));

var ExifCompression;
(function (ExifCompression) {
    ExifCompression[ExifCompression["Uncompressed"] = 1] = "Uncompressed";
    ExifCompression[ExifCompression["CCITT"] = 2] = "CCITT";
    ExifCompression[ExifCompression["T4Group3Fax"] = 3] = "T4Group3Fax";
    ExifCompression[ExifCompression["T6Group3Fax"] = 4] = "T6Group3Fax";
    ExifCompression[ExifCompression["LZW"] = 5] = "LZW";
    ExifCompression[ExifCompression["JpgOldStyle"] = 6] = "JpgOldStyle";
    ExifCompression[ExifCompression["Jpg"] = 7] = "Jpg";
    ExifCompression[ExifCompression["AdobeDeflate"] = 8] = "AdobeDeflate";
    ExifCompression[ExifCompression["JBigBw"] = 9] = "JBigBw";
    ExifCompression[ExifCompression["JBigColor"] = 10] = "JBigColor";
    ExifCompression[ExifCompression["JpegAlt"] = 99] = "JpegAlt";
    ExifCompression[ExifCompression["Kodak262"] = 262] = "Kodak262";
    ExifCompression[ExifCompression["Next"] = 32766] = "Next";
    ExifCompression[ExifCompression["SonyRawCompressed"] = 32767] = "SonyRawCompressed";
    ExifCompression[ExifCompression["PackedRaw"] = 32769] = "PackedRaw";
    ExifCompression[ExifCompression["SamsungSrwCompressed"] = 32770] = "SamsungSrwCompressed";
    ExifCompression[ExifCompression["CCIRLEW"] = 32771] = "CCIRLEW";
    ExifCompression[ExifCompression["SamsungSrwCompressed2"] = 32772] = "SamsungSrwCompressed2";
    ExifCompression[ExifCompression["Packbits"] = 32773] = "Packbits";
    ExifCompression[ExifCompression["Thunderscan"] = 32809] = "Thunderscan";
    ExifCompression[ExifCompression["KodakKdcCompressed"] = 32867] = "KodakKdcCompressed";
    ExifCompression[ExifCompression["IT8CTPAD"] = 32895] = "IT8CTPAD";
    ExifCompression[ExifCompression["IT8LW"] = 32896] = "IT8LW";
    ExifCompression[ExifCompression["IT8MP"] = 32897] = "IT8MP";
    ExifCompression[ExifCompression["IT8BL"] = 32898] = "IT8BL";
    ExifCompression[ExifCompression["PixarFilm"] = 32908] = "PixarFilm";
    ExifCompression[ExifCompression["PixarLog"] = 32909] = "PixarLog";
    ExifCompression[ExifCompression["Deflate"] = 32946] = "Deflate";
    ExifCompression[ExifCompression["DCS"] = 32947] = "DCS";
    ExifCompression[ExifCompression["AperioJpeg2000YCbCr"] = 33003] = "AperioJpeg2000YCbCr";
    ExifCompression[ExifCompression["AperioJpeg2000RGB"] = 33005] = "AperioJpeg2000RGB";
    ExifCompression[ExifCompression["JBig"] = 34661] = "JBig";
    ExifCompression[ExifCompression["SGILog"] = 34676] = "SGILog";
    ExifCompression[ExifCompression["SGILog24"] = 34677] = "SGILog24";
    ExifCompression[ExifCompression["Jpeg2000"] = 34712] = "Jpeg2000";
    ExifCompression[ExifCompression["NikonNEFCompressed"] = 34713] = "NikonNEFCompressed";
    ExifCompression[ExifCompression["JBig2TiffFx"] = 34715] = "JBig2TiffFx";
    ExifCompression[ExifCompression["MicrosoftBinaryLevelCodec"] = 34718] = "MicrosoftBinaryLevelCodec";
    ExifCompression[ExifCompression["MicrosoftProgressiveTransformCodec"] = 34719] = "MicrosoftProgressiveTransformCodec";
    ExifCompression[ExifCompression["MicrosoftVector"] = 34720] = "MicrosoftVector";
    ExifCompression[ExifCompression["ESRCLerc"] = 34887] = "ESRCLerc";
    ExifCompression[ExifCompression["LossyJpeg"] = 34892] = "LossyJpeg";
    ExifCompression[ExifCompression["LZMA2"] = 34925] = "LZMA2";
    ExifCompression[ExifCompression["Zstd"] = 34926] = "Zstd";
    ExifCompression[ExifCompression["WepP"] = 34927] = "WepP";
    ExifCompression[ExifCompression["PNG"] = 34933] = "PNG";
    ExifCompression[ExifCompression["JpegXR"] = 34934] = "JpegXR";
    ExifCompression[ExifCompression["KodakDCRCompressed"] = 65000] = "KodakDCRCompressed";
    ExifCompression[ExifCompression["PentaxPEFCompressed"] = 65535] = "PentaxPEFCompressed";
})(ExifCompression || (ExifCompression = {}));
var ExifLightSource;
(function (ExifLightSource) {
    ExifLightSource[ExifLightSource["Unknown"] = 0] = "Unknown";
    ExifLightSource[ExifLightSource["Daylight"] = 1] = "Daylight";
    ExifLightSource[ExifLightSource["Flourescent"] = 2] = "Flourescent";
    ExifLightSource[ExifLightSource["Tungsten"] = 3] = "Tungsten";
    ExifLightSource[ExifLightSource["Flash"] = 4] = "Flash";
    ExifLightSource[ExifLightSource["FineWeather"] = 5] = "FineWeather";
    ExifLightSource[ExifLightSource["Cloudy"] = 6] = "Cloudy";
    ExifLightSource[ExifLightSource["Shade"] = 7] = "Shade";
    ExifLightSource[ExifLightSource["DaylightFlourescent"] = 8] = "DaylightFlourescent";
    ExifLightSource[ExifLightSource["DayWhiteFlourescent"] = 9] = "DayWhiteFlourescent";
    ExifLightSource[ExifLightSource["CoolWhiteFlourescent"] = 10] = "CoolWhiteFlourescent";
    ExifLightSource[ExifLightSource["WhiteFlourescent"] = 11] = "WhiteFlourescent";
    ExifLightSource[ExifLightSource["WarmWhiteFlourescent"] = 12] = "WarmWhiteFlourescent";
    ExifLightSource[ExifLightSource["StandardLightA"] = 13] = "StandardLightA";
    ExifLightSource[ExifLightSource["StandardLightB"] = 14] = "StandardLightB";
    ExifLightSource[ExifLightSource["StandardLightC"] = 15] = "StandardLightC";
    ExifLightSource[ExifLightSource["D55"] = 16] = "D55";
    ExifLightSource[ExifLightSource["D65"] = 17] = "D65";
    ExifLightSource[ExifLightSource["D75"] = 18] = "D75";
    ExifLightSource[ExifLightSource["D50"] = 19] = "D50";
    ExifLightSource[ExifLightSource["ISOStudioTungten"] = 20] = "ISOStudioTungten";
    ExifLightSource[ExifLightSource["Other"] = 255] = "Other";
})(ExifLightSource || (ExifLightSource = {}));
var ExifFlashValues;
(function (ExifFlashValues) {
    ExifFlashValues[ExifFlashValues["NoFlash"] = 0] = "NoFlash";
    ExifFlashValues[ExifFlashValues["Fired"] = 1] = "Fired";
    ExifFlashValues[ExifFlashValues["FiredReturnNotDetected"] = 5] = "FiredReturnNotDetected";
    ExifFlashValues[ExifFlashValues["FiredReturnDetected"] = 7] = "FiredReturnDetected";
    ExifFlashValues[ExifFlashValues["OnDidNotFire"] = 8] = "OnDidNotFire";
    ExifFlashValues[ExifFlashValues["OnFired"] = 9] = "OnFired";
    ExifFlashValues[ExifFlashValues["OnReturnNotDetected"] = 13] = "OnReturnNotDetected";
    ExifFlashValues[ExifFlashValues["OnReturnDetected"] = 15] = "OnReturnDetected";
    ExifFlashValues[ExifFlashValues["OffDidNotFire"] = 16] = "OffDidNotFire";
    ExifFlashValues[ExifFlashValues["OffDidNotFireReturnNotDetected"] = 20] = "OffDidNotFireReturnNotDetected";
    ExifFlashValues[ExifFlashValues["AutoDidNotFire"] = 24] = "AutoDidNotFire";
    ExifFlashValues[ExifFlashValues["AutoFired"] = 25] = "AutoFired";
    ExifFlashValues[ExifFlashValues["AutoFiredReturnNotDetected"] = 29] = "AutoFiredReturnNotDetected";
    ExifFlashValues[ExifFlashValues["AutoFiredReturnDetected"] = 31] = "AutoFiredReturnDetected";
    ExifFlashValues[ExifFlashValues["NoFlashFunction"] = 32] = "NoFlashFunction";
    ExifFlashValues[ExifFlashValues["OffNoFlashFunction"] = 48] = "OffNoFlashFunction";
    ExifFlashValues[ExifFlashValues["FiredRedEyeReduction"] = 65] = "FiredRedEyeReduction";
    ExifFlashValues[ExifFlashValues["FiredRedEyeReductionReturnNotDetected"] = 69] = "FiredRedEyeReductionReturnNotDetected";
    ExifFlashValues[ExifFlashValues["FiredRedEyeReductionReturnDetected"] = 71] = "FiredRedEyeReductionReturnDetected";
    ExifFlashValues[ExifFlashValues["OnRedEyeReduction"] = 73] = "OnRedEyeReduction";
    ExifFlashValues[ExifFlashValues["OnRedEyeReductionReturnNotDetected"] = 77] = "OnRedEyeReductionReturnNotDetected";
    ExifFlashValues[ExifFlashValues["OnRedEyeReductionReturnDetected"] = 79] = "OnRedEyeReductionReturnDetected";
    ExifFlashValues[ExifFlashValues["OffRedEyeReduction"] = 80] = "OffRedEyeReduction";
    ExifFlashValues[ExifFlashValues["AutoDidNotFireRedEyeReduction"] = 88] = "AutoDidNotFireRedEyeReduction";
    ExifFlashValues[ExifFlashValues["AutoFiredRedEyeReduction"] = 89] = "AutoFiredRedEyeReduction";
    ExifFlashValues[ExifFlashValues["AutoFiredRedEyeReductionNotDetected"] = 89] = "AutoFiredRedEyeReductionNotDetected";
    ExifFlashValues[ExifFlashValues["AutoFiredRedEyeReductionDetected"] = 93] = "AutoFiredRedEyeReductionDetected";
})(ExifFlashValues || (ExifFlashValues = {}));
var ExifPreviewColorSpace;
(function (ExifPreviewColorSpace) {
    ExifPreviewColorSpace[ExifPreviewColorSpace["Unknown"] = 0] = "Unknown";
    ExifPreviewColorSpace[ExifPreviewColorSpace["GrayGamma22"] = 1] = "GrayGamma22";
    ExifPreviewColorSpace[ExifPreviewColorSpace["sRGB"] = 2] = "sRGB";
    ExifPreviewColorSpace[ExifPreviewColorSpace["AdobeRGB"] = 3] = "AdobeRGB";
    ExifPreviewColorSpace[ExifPreviewColorSpace["ProPhotoRGB"] = 4] = "ProPhotoRGB";
})(ExifPreviewColorSpace || (ExifPreviewColorSpace = {}));
var ExifEmbedPolicy;
(function (ExifEmbedPolicy) {
    ExifEmbedPolicy[ExifEmbedPolicy["AllowCopying"] = 0] = "AllowCopying";
    ExifEmbedPolicy[ExifEmbedPolicy["EmbedIfUsed"] = 1] = "EmbedIfUsed";
    ExifEmbedPolicy[ExifEmbedPolicy["NeverEmbed"] = 2] = "NeverEmbed";
    ExifEmbedPolicy[ExifEmbedPolicy["NoRestrictions"] = 3] = "NoRestrictions";
})(ExifEmbedPolicy || (ExifEmbedPolicy = {}));
var ExifSubjectDistance;
(function (ExifSubjectDistance) {
    ExifSubjectDistance[ExifSubjectDistance["Unknown"] = 0] = "Unknown";
    ExifSubjectDistance[ExifSubjectDistance["Macro"] = 1] = "Macro";
    ExifSubjectDistance[ExifSubjectDistance["Close"] = 2] = "Close";
    ExifSubjectDistance[ExifSubjectDistance["Distant"] = 3] = "Distant";
})(ExifSubjectDistance || (ExifSubjectDistance = {}));
var ExifSharpness;
(function (ExifSharpness) {
    ExifSharpness[ExifSharpness["Normal"] = 0] = "Normal";
    ExifSharpness[ExifSharpness["Soft"] = 1] = "Soft";
    ExifSharpness[ExifSharpness["Hard"] = 2] = "Hard";
})(ExifSharpness || (ExifSharpness = {}));
var ExifSceneCaptureType;
(function (ExifSceneCaptureType) {
    ExifSceneCaptureType[ExifSceneCaptureType["Standard"] = 0] = "Standard";
    ExifSceneCaptureType[ExifSceneCaptureType["Landscape"] = 1] = "Landscape";
    ExifSceneCaptureType[ExifSceneCaptureType["Portrait"] = 2] = "Portrait";
    ExifSceneCaptureType[ExifSceneCaptureType["Night"] = 3] = "Night";
    ExifSceneCaptureType[ExifSceneCaptureType["Other"] = 4] = "Other";
})(ExifSceneCaptureType || (ExifSceneCaptureType = {}));
var ExifGainControl;
(function (ExifGainControl) {
    ExifGainControl[ExifGainControl["None"] = 0] = "None";
    ExifGainControl[ExifGainControl["LowGainUp"] = 1] = "LowGainUp";
    ExifGainControl[ExifGainControl["HighGainUp"] = 2] = "HighGainUp";
    ExifGainControl[ExifGainControl["LowGainDown"] = 3] = "LowGainDown";
    ExifGainControl[ExifGainControl["HighGainDown"] = 4] = "HighGainDown";
})(ExifGainControl || (ExifGainControl = {}));
var ExifContrast;
(function (ExifContrast) {
    ExifContrast[ExifContrast["Normal"] = 0] = "Normal";
    ExifContrast[ExifContrast["Low"] = 1] = "Low";
    ExifContrast[ExifContrast["High"] = 2] = "High";
})(ExifContrast || (ExifContrast = {}));
var ExifSaturation;
(function (ExifSaturation) {
    ExifSaturation[ExifSaturation["Normal"] = 0] = "Normal";
    ExifSaturation[ExifSaturation["Low"] = 1] = "Low";
    ExifSaturation[ExifSaturation["High"] = 2] = "High";
})(ExifSaturation || (ExifSaturation = {}));

/**
 * Ways in which missing data can be treated.
 *
 * [Official Docs](https://docs.aws.amazon.com/AmazonCloudWatch/latest/monitoring/AlarmThatSendsEmail.html#alarms-and-missing-data)
 */
var StepFunctionMissingDataTreatment;
(function (StepFunctionMissingDataTreatment) {
    /** The alarm does not consider missing data points when evaluating whether to change state. */
    StepFunctionMissingDataTreatment["missing"] = "missing";
    /** The current alarm state is maintained. */
    StepFunctionMissingDataTreatment["ignore"] = "ignore";
    /** Missing data points are treated as breaching the threshold. */
    StepFunctionMissingDataTreatment["breaching"] = "breaching";
    /** Missing data points are treated as being within the threshold. */
    StepFunctionMissingDataTreatment["notBreaching"] = "notBreaching";
})(StepFunctionMissingDataTreatment || (StepFunctionMissingDataTreatment = {}));

async function networkDelay(delay) {
    let waitFor;
    if (typeof delay === 'number') {
        waitFor = delay;
    }
    else if (Array.isArray(delay)) {
        waitFor = between(...delay);
    }
    else {
        const lookup = {
            [NetworkDelay.lifi]: [10, 10000],
            [NetworkDelay.mobile2g]: [300, 1200],
            [NetworkDelay.mobile3g]: [200, 750],
            [NetworkDelay.mobile4g]: [75, 200],
            [NetworkDelay.wifi]: [25, 75],
            [NetworkDelay.lazer]: [1, 5],
        };
        waitFor = between(...lookup[delay]);
    }
    await wait(waitFor);
}

function normalizeRef(r) {
    r = r.replace('/', '.');
    r = r.slice(0, 1) === '.' ? r.slice(1) : r;
    return r;
}
function parts(r) {
    return normalizeRef(r).split('.');
}
/**
 * return the last component of the path
 * which typically would represent the 'id'
 * of a list-node
 */
function leafNode(r) {
    return parts(r).pop();
}
/**
 * Joins a set of paths together and converts into
 * correctly formatted "dot notation" directory path
 */
function join(...paths) {
    return paths
        .map((p) => {
        return p.replace(/[/\\]/gm, '.');
    })
        .map((p) => (p.slice(-1) === '.' ? p.slice(0, p.length - 1) : p))
        .map((p) => (p.slice(0, 1) === '.' ? p.slice(1) : p))
        .join('.');
}
/**
 * Given a path, returns the parent path and child key
 */
function keyAndParent(dotPath) {
    const sections = dotPath.split('.');
    const changeKey = sections.pop();
    const parent = sections.join('.');
    return { parent, key: changeKey };
}
/** Get the parent DB path */
function getParent(dotPath) {
    return keyAndParent(dotPath).parent;
}
/** Get the Key from the end of a path string */
function getKey(dotPath) {
    return keyAndParent(dotPath).key;
}
function stripLeadingDot(str) {
    return str.slice(0, 1) === '.' ? str.slice(1) : str;
}
function removeDots(str) {
    return str ? str.replace(/\./g, '') : undefined;
}

/**
 * Sets a value at a nested point within base object passed in. This is meant as a
 * replacement to use of `lodash.set()`.
 *
 * @param obj the base object which is being mutated
 * @param dotPath the path into the object where the mutation will take place, delimited by `.`
 * @param value The value to set at the _dotPath_
 * @param createIfNonExistant by default, if the path to the object does not exist then an error is thrown but if you want you can state the desire to have the full path created
 */
function set(obj, dotPath, value, createIfNonExistant = true) {
    if (!dotPath && !createIfNonExistant) {
        throw new FireMockError(`Attempt to set value into a dotPath but the dotPath was empty!`, 'not-allowed');
    }
    else if (!dotPath && createIfNonExistant && typeof value === 'object') {
        Object.entries(value).forEach(([key, _value]) => {
            set(obj, key, _value);
        });
        return;
    }
    const parts = dotPath === null || dotPath === void 0 ? void 0 : dotPath.split(/\??\./);
    const allButLast = (parts === null || parts === void 0 ? void 0 : parts.slice(0, parts.length - 1)) || Object.keys(value);
    const key = parts.pop();
    let ref = obj;
    // iterate the ref to the leaf node
    allButLast.forEach((p) => {
        if (!ref[p]) {
            if (createIfNonExistant) {
                ref[p] = {};
            }
            else {
                throw new FireMockError(`The dotPath -- ${dotPath} -- does not exist in the passed in object. You must either expressly state that you want the object structure created or this a real error that must be addressed otherwise. The part of the path which this failed on was "${p}".`, 'not-allowed');
            }
        }
        else if (typeof ref[p] !== 'object') {
            throw new FireMockError(`Failed to set the path of "${dotPath}" of the passed in base object because the base object had a scalar value along that path and setting this would have changed the object's data structure in way which is not allowed! The scalar value was found in the "${p}" component of the path.`, 'not-allowed');
        }
        ref = ref[p];
    });
    ref[key] = value;
}

const orderByChild = (child) => {
    return (a, b) => {
        return a[child] > b[child] ? 1 : a[child] === b[child] ? 0 : -1;
    };
};
const orderByKey$1 = (a, b) => {
    return a.id > b.id ? -1 : a.id === b.id ? 0 : 1;
};
const orderByValue$1 = (a, b) => {
    return a.value > b.value ? -1 : a.value === b.value ? 0 : 1;
};
function isOrderByChild(query, fn) {
    return query.identity.orderBy === RtdbOrder.orderByChild;
}

var sortFns = /*#__PURE__*/Object.freeze({
  __proto__: null,
  orderByChild: orderByChild,
  orderByKey: orderByKey$1,
  orderByValue: orderByValue$1,
  isOrderByChild: isOrderByChild
});

class Auth {
    constructor() {
        throw new Error('You should not call this constructor directly! Instead use the auth() accessor to get to this API.');
    }
}

const data = {
    providerId: 'emailPassword',
    signInMethod: 'emailPassword',
};
/**
 * **EmailAuthProvider** API mocking. Details on the API can be found
 * here: https://firebase.google.com/docs/reference/js/firebase.auth.EmailAuthProvider
 */
class EmailAuthProvider {
    constructor() {
        this.providerId = data.providerId;
    }
    /**
     * Produces a `credential` to a user account (typically an anonymous account)
     * which can then be linked to the account using `linkWithCredential`.
     */
    static credential(email, password) {
        return {
            ...data,
            toJSON() {
                return JSON.stringify(data);
            },
        };
    }
    /**
     * Initialize an EmailAuthProvider credential using an email and an email link after
     * a sign in with email link operation.
     */
    static credentialWithLink(email, emailLink) {
        return {
            ...data,
            toJSON() {
                return JSON.stringify(data);
            },
        };
    }
}
EmailAuthProvider.PROVIDER_ID = 'emailPassword';

class FacebookAuthProvider {
    constructor() {
        this.providerId = 'facebook';
    }
    static credential(token) {
        throw new Error('FacebookAuthProvider not implemented yet');
    }
    addScope(scope) {
        throw new Error('not implemented');
    }
    setCustomParameters(params) {
        throw new Error('not implemented');
    }
}

class GithubAuthProvider {
    static credential(idToken, accessToken) {
        throw new Error('not implemented');
    }
    addScope(scope) {
        throw new Error('not implemented');
    }
    setCustomParameters(params) {
        throw new Error('not implemented');
    }
}
GithubAuthProvider.PROVIDER_ID = 'github';

class GoogleAuthProvider {
    static credential(idToken, accessToken) {
        throw new Error('not implemented');
    }
    addScope(scope) {
        throw new Error('not implemented');
    }
    setCustomParameters(params) {
        throw new Error('not implemented');
    }
}
GoogleAuthProvider.PROVIDER_ID = 'google';

class TwitterAuthProvider {
    constructor() {
        this.providerId = 'twitter';
    }
    static credential(idToken, accessToken) {
        throw new Error('not implemented');
    }
    addScope(scope) {
        throw new Error('not implemented');
    }
    setCustomParameters(params) {
        throw new Error('not implemented');
    }
}
TwitterAuthProvider.PROVIDER_ID = 'twitter';

class SAMLAuthProvider {
    constructor() {
        this.providerId = 'saml';
    }
}

class OAuthProvider {
    // tslint:disable-next-line: no-empty
    constructor(providerId) {
        this.providerId = 'oauth';
        this.providerId = providerId;
    }
    addScope(_scope) {
        throw new Error('not implemented');
    }
    credential(_idToken, _accessToken) {
        throw new Error('not implemented');
    }
    // tslint:disable-next-line: ban-types
    setCustomParameters(_customOAuthParameters) {
        throw new Error('not implemented');
    }
}

class PhoneAuthProvider {
    constructor() {
        this.providerId = 'phone';
    }
    static credential(_verificationId, _verificationCode) {
        throw new Error('not implemented');
    }
    verifyPhoneNumber(_phoneNumber, _applicationVerifier) {
        throw new Error('not-implemented');
    }
}
PhoneAuthProvider.PROVIDER_ID = 'phone';

class RecaptchaVerifier {
    clear() {
        // TODO:
    }
    render() {
        throw new Error('not-implemented');
    }
    verify() {
        throw new Error('not-implemented');
    }
}

/**
 * The mocked Auth API for a Client SDK
 */
const authProviders = {
    Auth: Auth,
    EmailAuthProvider,
    FacebookAuthProvider,
    GithubAuthProvider,
    GoogleAuthProvider,
    TwitterAuthProvider,
    SAMLAuthProvider,
    OAuthProvider,
    PhoneAuthProvider,
    RecaptchaVerifier,
};
const fn = () => {
    throw new Error('not allowed');
};
var _authProviders = (authProviders || fn);

function getIdToken(api) {
    return async () => {
        const user = api.getCurrentUser();
        await networkDelay(NetworkDelay.wifi);
        if (!user) {
            throw new Error('not logged in');
        }
        if (user.tokenIds) {
            return atRandom(user.tokenIds);
        }
        else {
            return Math.random().toString(36).substr(2, 10);
        }
    };
}

function getIdTokenResult(api) {
    return async (forceRefresh) => {
        return {
            authTime: '',
            claims: {},
            expirationTime: '',
            issuedAtTime: '',
            signInProvider: '',
            signInSecondFactor: '',
            token: await getIdToken(api)(),
        };
    };
}

const notImplemented$2 = {
    /** Deletes and signs out the user. */
    delete() {
        throw new Error('the Mock Auth feature for delete() is not yet implemented');
    },
    linkAndRetrieveDataWithCredential(credential) {
        throw new Error(`linkAndRetrieveDataWithCredential() is not implemented yet in the client-sdk's mock auth`);
    },
    linkWithCredential(credential) {
        throw new Error(`linkWithCredential() is not implemented yet in the client-sdk's mock auth`);
    },
    async linkWithPhoneNumber(phoneNUmber, applicationVerificer) {
        return Promise.resolve(fakeApplicationVerifier);
    },
    linkWithPopup(provider) {
        throw new Error(`linkWithPopup() is not implemented yet in the client-sdk's mock auth`);
    },
    linkWithRedirect(provider) {
        throw new Error(`linkWithRedirect() is not implemented yet in the client-sdk's mock auth`);
    },
    reauthenticateAndRetrieveDataWithCredential(credential) {
        throw new Error(`reauthenticateAndRetrieveDataWithCredential() is not implemented yet in the client-sdk's mock auth`);
    },
    reauthenticateWithCredential(credential) {
        throw new Error(`reauthenticateWithCredential() is not implemented yet in the client-sdk's mock auth`);
    },
    async reauthenticateWithPhoneNumber(phoneNumber, applicationVerifier) {
        return Promise.resolve(fakeApplicationVerifier);
    },
    reauthenticateWithPopup(provider) {
        throw new Error(`reauthenticateWithPopup() is not implemented yet in the client-sdk's mock auth`);
    },
    reauthenticateWithRedirect(provider) {
        throw new Error(`reauthenticateWithRedirect() is not implemented yet in the client-sdk's mock auth`);
    },
    reload() {
        return;
    },
    sendEmailVerification(actionCodeSettings) {
        throw new Error(`sendEmailVerification() is not implemented yet in the client-sdk's mock auth`);
    },
    toJSON() {
        return {};
    },
    unlink(provider) {
        throw new Error(`unlink() is not implemented yet in the client-sdk's mock auth`);
    },
    updatePhoneNumber(phoneCredential) {
        throw new Error(`updatePhoneNumber() is not implemented yet in the client-sdk's mock auth`);
    },
};

/**
 * **updateEmail**
 *
 * Gives a logged in user the ability to update their email address.
 *
 * Possible Errors:
 *
 * - auth/invalid-email
 * - auth/email-already-in-use
 * - auth/requires-recent-login
 *
 * [Documentation](https://firebase.google.com/docs/reference/js/firebase.User#update-email)
 *
 * > Note: The `forceLogin` is not part of Firebase API but allows mock user to force the
 * error condition associated with a user who's been logged in a long time.
 */
const updateEmail = (api) => async (newEmail, forceLogin) => {
    if (forceLogin) {
        throw new FireMockError("updating a user's email address requires that the user have recently logged in; use 'reauthenticateWithCredential' to address this error.", 'auth/requires-recent-login');
    }
    await networkDelay(NetworkDelay.wifi);
    api.updateUser(api.getCurrentUser(), { email: newEmail });
};

/**
 * **updatePassword**
 *
 * Updates the password for a logged in user. For security reasons, this operations
 * requires the user to have recently signed in.
 *
 * Errors:
 *
 * - auth/weak-password
 * - auth/required-recent-login ( can use `reauthenticateWithCredential` to resolve this )
 *
 * > Note: the `notRecentLogin` is NOT part of the normal API but allows mock users to force
 * the `auth/required-recent-login` error.
 *
 * [Docs](https://firebase.google.com/docs/reference/js/firebase.User#updatepassword)
 */
const updatePassword = (api) => async (newPassword, notRecentLogin) => {
    if (notRecentLogin) {
        throw new FireMockError("updating a user's password requires that the user have recently logged in; use 'reauthenticateWithCredential' to address this error.", 'auth/required-recent-login');
    }
    await networkDelay(NetworkDelay.wifi);
    api.updateUser(api.getCurrentUser(), {
        password: newPassword,
    });
};

/**
 * Updates a user's profile data.
 *
 * Example:
 *
 * ```typescript
 * user.updateProfile({
      displayName: "Jane Q. User",
      photoURL: "https://example.com/jane-q-user/profile.jpg"
  })
 * ```

 [Documentation](https://firebase.google.com/docs/reference/js/firebase.User#updateprofile)
 */
const updateProfile = (api) => async (profile) => {
    await networkDelay(NetworkDelay.wifi);
    api.updateUser(api.getCurrentUser(), profile);
};

/**
 * Combines the property and API aspects into a `User` object where
 * the API aspects are provided access to the `IMockAuthMgmt` api to
 * ensure function calls effect the correct AUTH instance.
 */
function createUser(api, user) {
    const fns = {
        ...notImplemented$2,
        getIdToken: getIdToken(api),
        getIdTokenResult: getIdTokenResult(api),
        updateEmail: updateEmail(api),
        updatePassword: updatePassword(api),
        updateProfile: updateProfile(api),
    };
    return {
        isAnonymous: false,
        emailVerified: user.email ? true : false,
        ...fns,
        ...user,
    };
}

let _currentUser = null;
const toMockUser = (user, metadata) => {
    return {
        kind: 'MockUserRecord',
        uid: user.uid,
        email: user.email,
        emailVerified: user.emailVerified,
        disabled: false,
        customClaims: {},
        password: user.password,
        metadata,
        providerData: undefined,
        toJSON: () => ({}),
    };
};
/**
 * Creates an Auth Admin Manager API which can be used for advanced
 * test use cases. This API will be appropriately adjusted based on
 * whether this is an Admin or Client SDK.
 */
function createAuthManager(sdk) {
    const _anonymousUidQueue = [];
    let _providers;
    const _authObservers = [];
    let _knownUsers = [];
    let _networkDelay;
    const getAuthObservers = () => {
        return _authObservers;
    };
    const addAuthObserver = (ob) => {
        _authObservers.push(ob);
    };
    const hasProvider = (provider) => {
        return _providers.includes(provider);
    };
    const addProvider = (provider) => {
        if (_providers.includes(provider)) {
            throw new FireMockError(`Failed adding "${provider}" as a new Firemock auth provider as it was already listed as a provider!`, 'already-exists');
        }
        _providers.push(provider);
    };
    const removeProvider = (provider) => {
        if (!_providers.includes(provider)) {
            throw new FireMockError(`Failed to remove "${provider}" as a Firemock auth provider as it was not listed as a provider!`, 'does-not-exists');
        }
        _providers = _providers.filter((i) => i !== provider);
    };
    const findKnownUser = (prop, value) => {
        return _knownUsers.find((u) => u[prop] === value);
    };
    const getCurrentUser = () => {
        const u = _currentUser ? findKnownUser('uid', _currentUser) : undefined;
        return u;
    };
    const setCurrentUser = (user, force) => {
        var _a;
        let lookingFor;
        if (!user) {
            _currentUser = undefined;
        }
        else if (typeof user === 'string') {
            lookingFor = user;
        }
        else if (isUser(user) || isMockUserRecord(user)) {
            lookingFor = user.uid;
        }
        else if (isUserCredential(user)) {
            lookingFor = user.user.uid;
        }
        const u = findKnownUser('uid', lookingFor);
        const uid = u ? u.uid : undefined;
        if (!uid && !force) {
            throw new FireMockError(`Attempt to set current user failed as UID [${lookingFor}] could not be found in the known users being managed by the Auth service!`, 'user-missing');
        }
        if (_currentUser !== uid || force) {
            _currentUser = uid;
            // eslint-disable-next-line @typescript-eslint/no-unsafe-return
            (_a = getAuthObservers()) === null || _a === void 0 ? void 0 : _a.map((o) => {
                const currentUser = getCurrentUser();
                return o(currentUser ? createUser(authManager, toUser(currentUser)) : undefined);
            });
        }
        return u;
    };
    const login = (user) => {
        if (!user) {
            throw new FireMockError(`Attempt to login without passing a valid user representation!`, 'not-allowed');
        }
        return createUser(authManager, toUser(setCurrentUser(user)));
    };
    const logout = () => {
        if (_providers) {
            if (_providers.includes('anonymous')) {
                setCurrentUser(getAnonymousUid());
            }
            else {
                setCurrentUser(undefined, true);
            }
        }
    };
    const createNewUid = () => {
        return `${Math.random().toString(36).substr(2, 10)}`;
    };
    const getAnonymousUid = () => {
        if (_anonymousUidQueue.length > 0) {
            const uid = _anonymousUidQueue.shift();
            return uid;
        }
        else {
            return `anonymous-user-${createNewUid()}`;
        }
    };
    const addToUserPool = (user) => {
        let mockUser;
        // for those that need fake data:
        const metadata = {
            lastSignInTime: undefined,
            creationTime: new Date().toUTCString(),
            toJSON: () => ({}),
        };
        if (isMockUserRecord(user)) {
            mockUser = user;
        }
        else if (isUserRecord(user)) {
            mockUser = {
                password: '123456789',
                ...user,
                kind: 'MockUserRecord',
            };
        }
        else if (isUserCredential(user)) {
            mockUser = {
                kind: 'MockUserRecord',
                uid: user.user.uid,
                emailVerified: user.user.emailVerified,
                disabled: false,
                customClaims: {},
                phoneNumber: user.user.phoneNumber,
                isAnonymous: user.user.isAnonymous,
                photoURL: user.user.photoURL,
                tokenIds: [],
                metadata,
                providerData: undefined,
                toJSON: () => ({}),
            };
        }
        else if (isUser(user)) {
            mockUser = toMockUser(user, metadata);
        }
        else if ('email' in user) {
            mockUser = {
                kind: 'MockUserRecord',
                uid: user.uid || uuid(),
                emailVerified: true,
                disabled: false,
                customClaims: {},
                tokenIds: [],
                metadata,
                providerData: undefined,
                ...user,
            };
        }
        _knownUsers.push(mockUser);
        return _knownUsers.length;
    };
    const removeFromUserPool = (uid) => {
        _knownUsers = _knownUsers.filter((i) => i.uid !== uid);
        return _knownUsers.length;
    };
    const knownUsers = () => {
        return _knownUsers;
    };
    const initializeAuth = async (config) => {
        const fullConfig = {
            providers: (config === null || config === void 0 ? void 0 : config.providers) || ['anonymous'],
            users: (config === null || config === void 0 ? void 0 : config.users) || [],
            options: {
                networkDelay: NetworkDelay.wifi,
                ...config === null || config === void 0 ? void 0 : config.options,
            },
        };
        _providers =
            typeof fullConfig.providers === 'function'
                ? await fullConfig.providers()
                : fullConfig.providers;
        setNetworkDelay(fullConfig.options.networkDelay);
        const users = typeof fullConfig.users === 'function'
            ? await fullConfig.users()
            : fullConfig.users;
        _knownUsers = [];
        users === null || users === void 0 ? void 0 : users.forEach((u) => {
            addToUserPool(u);
        });
    };
    const updateUser = (user, updates) => {
        let u;
        if (!isMockUserRecord(updates)) {
            updates = {
                // ...(updates.disabled !== undefined
                //   ? { disabled: updates.disabled }
                //   : {}),
                ...(updates.displayName !== undefined
                    ? { displayName: updates.displayName }
                    : {}),
                ...(updates.email !== undefined ? { email: updates.email } : {}),
                ...(updates.emailVerified !== undefined
                    ? { emailVerified: updates.emailVerified }
                    : {}),
                // ...(updates.password !== undefined
                //   ? { password: updates.password }
                //   : {}),
                ...(updates.photoURL !== undefined
                    ? { photoURL: updates.photoURL }
                    : {}),
                ...(updates.multiFactor !== undefined
                    ? { clientMultiFactor: { enrolledFactors: [] } }
                    : {}),
            };
        }
        if (typeof user === 'string') {
            u = findKnownUser('uid', user);
            if (!u) {
                throw new FireMockError(`Attempt to update the user with UID of "${user}" failed because this user is not defined in the mock Auth instance!`);
            }
        }
        else {
            u = user;
        }
        _knownUsers = _knownUsers.map((i) => i.uid === u.uid ? { ...i, ...updates } : i);
    };
    const queueAnonymousUidPool = () => {
        //
    };
    const authProviders = isAdminSdk(sdk)
        ? undefined
        : _authProviders;
    const getAuthProvidersNames = () => {
        return _providers;
    };
    const networkDelay$1 = async () => {
        await networkDelay(_networkDelay);
    };
    const setNetworkDelay = (delay) => {
        _networkDelay = delay;
    };
    const authManager = {
        getAuthObservers,
        addAuthObserver,
        getCurrentUser,
        setCurrentUser,
        login,
        logout,
        createNewUid,
        getAnonymousUid,
        findKnownUser,
        hasProvider,
        updateUser,
        addToUserPool,
        knownUsers,
        initializeAuth,
        queueAnonymousUidPool,
        addProvider,
        removeProvider,
        networkDelay: networkDelay$1,
        removeFromUserPool,
        setNetworkDelay,
        authProviders,
        getAuthProvidersNames,
    };
    return authManager;
}

/**
 * Passing the in the management API returns the Firebase Client Auth API
 */
const createClientAuth = (api) => {
    return {
        ...notImplemented$1,
        ...implemented$1(api),
    };
};

const implemented$1 = (api) => ({
    tenantId: '',
    languageCode: '',
    settings: {
        appVerificationDisabledForTesting: false,
    },
    app: {
        name: 'mocked-app',
        options: {},
        async delete() {
            return;
        },
        automaticDataCollectionEnabled: false,
    },
    onAuthStateChanged(observer, error, completed) {
        api.addAuthObserver(observer);
        const currentUser = api.getCurrentUser();
        observer(currentUser ? createUser(api, toUser(currentUser)) : undefined);
        return undefined;
    },
    async setPersistence() {
        console.warn(`currently firemock accepts calls to setPersistence() but it doesn't support it.`);
        return;
    },
    signInAnonymously: async () => {
        await api.networkDelay();
        if (api.hasProvider('anonymous')) {
            const user = createUser(api, {
                isAnonymous: true,
                uid: api.getAnonymousUid(),
            });
            const credential = {
                signInMethod: 'anonymous',
                providerId: 'anonymous',
                toJSON: () => ({
                    signInMethod: 'anonymous',
                    providerId: 'anonymous',
                }),
            };
            const userCredentials = completeUserCredential({ user, credential });
            api.addToUserPool(userCredentials);
            api.setCurrentUser(userCredentials);
            return userCredentials;
        }
        else {
            throw new FireMockError('you must enable anonymous auth in the Firebase Console', 'auth/operation-not-allowed');
        }
    },
    async updateCurrentUser(updates) {
        api.updateUser(api.getCurrentUser().uid, updates);
    },
    /**
     * Sign into Firebase with Email and Password:
     * [Docs](https://firebase.google.com/docs/reference/js/firebase.auth.Auth#signinwithemailandpassword)
     *
     * Error Codes:
     *  - auth/invalid-email
     *  - auth/user-disabled
     *  - auth/user-not-found
     *  - auth/wrong-password
     */
    async signInWithEmailAndPassword(email, password) {
        await api.networkDelay();
        if (!emailValidationAllowed(api)()) {
            throw new FireMockError('email authentication not allowed', 'auth/operation-not-allowed');
        }
        if (!emailIsValidFormat()(email)) {
            throw new FireMockError(`invalid email: ${email}`, 'auth/invalid-email');
        }
        const user = api.findKnownUser('email', email);
        if (!user) {
            throw new FireMockError(`The email "${email}" is not a known user in the mock database`, `auth/user-not-found`);
        }
        if (user.disabled) {
            throw new FireMockError(`The user identified by "${user.email}" has been disabled!`, `auth/user-disabled`);
        }
        if (!emailHasCorrectPassword(api)(email, password)) {
            throw new FireMockError(`Invalid password for ${email}`, 'auth/wrong-password');
        }
        const partial = {
            user: {
                email: user.email,
                isAnonymous: false,
                emailVerified: user.emailVerified,
                uid: userUid(api)(email),
                displayName: user.displayName,
                ...createUser(api, user),
            },
            credential: {
                signInMethod: 'signInWithEmailAndPassword',
                providerId: '',
            },
            additionalUserInfo: {
                username: email,
            },
        };
        const u = completeUserCredential(partial);
        api.setCurrentUser(u);
        return u;
    },
    /**
     * Add a new user with the Email/Password provider.
     *
     * Possible errors:
     * - auth/email-already-in-use
     * - auth/invalid-email
     * - auth/operation-not-allowed
     * - auth/weak-password
     *
     * [Docs](https://firebase.google.com/docs/reference/js/firebase.auth.Auth#createuserwithemailandpassword)
     */
    async createUserWithEmailAndPassword(email, password) {
        await api.networkDelay();
        if (!emailValidationAllowed(api)()) {
            throw new FireMockError('email authentication not allowed', 'auth/operation-not-allowed');
        }
        if (emailExistsAsUserInAuth(api)(email)) {
            throw new FireMockError(`"${email}" user already exists`, 'auth/email-already-in-use');
        }
        if (!emailIsValidFormat()(email)) {
            throw new FireMockError(`"${email}" is not a valid email format`, 'auth/invalid-email');
        }
        const uid = userUid(api)(email);
        const partial = {
            user: {
                ...createUser(api, { email, uid }),
                isAnonymous: false,
                emailVerified: false,
            },
            credential: {
                signInMethod: 'signInWithEmailAndPassword',
                providerId: '',
            },
            additionalUserInfo: {
                username: email,
            },
        };
        const u = completeUserCredential(partial);
        api.addToUserPool({ uid: partial.user.uid, email, password });
        api.setCurrentUser(u);
        return u;
    },
    async confirmPasswordReset(code, newPassword) {
        return;
    },
    async sendPasswordResetEmail(email, actionCodeSetting) {
        console.group('Firemock: sendPasswordResetEmail()');
        console.log(`an email is being sent to "${email}".`);
        console.log(`the "action code settings" are:`, actionCodeSetting);
        console.groupEnd();
    },
    async signOut() {
        api.logout();
    },
    get currentUser() {
        return completeUserCredential({}).user;
    },
});

/**
 * Implements the interface's functions but functions just
 * throw errors.
 */
const notImplemented$1 = {
    useEmulator(url, opts) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    emulatorConfig: null,
    onAuthStateChanged(nextOrObserver) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInAnonymously() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async updateCurrentUser(user) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async confirmPasswordReset(code, newPassword) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async verifyPasswordResetCode(code) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async useDeviceLanguage() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async setPersistence() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    // async isSignInWithEmailLink(emailLink: string) {
    //   throw createError(
    //     'auth/not-implemented',
    //     'This feature is not implemented yet in FireMock auth module'
    //   );
    // },
    async signInWithEmailAndPassword(email, password) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async createUserWithEmailAndPassword(email, password) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async sendPasswordResetEmail(email, actionCodeSetting) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signOut() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async applyActionCode(code) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async checkActionCode(code) {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async fetchSignInMethodsForEmail() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async getRedirectResult() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    isSignInWithEmailLink() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    onIdTokenChanged() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async sendSignInLinkToEmail() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInAndRetrieveDataWithCredential() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithCredential() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithCustomToken() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithEmailLink() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithPhoneNumber() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithPopup() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    async signInWithRedirect() {
        throw createError('auth/not-implemented', 'This feature is not implemented yet in FireMock auth module');
    },
    // async useDeviceLanguage() {
    //   throw createError(
    //     'auth/not-implemented',
    //     'This feature is not implemented yet in FireMock auth module'
    //   );
    // },
    // async verifyPasswordResetCode(code: string) {
    //   throw createError(
    //     'auth/not-implemented',
    //     'This feature is not implemented yet in FireMock auth module'
    //   );
    // },
};

const emailExistsAsUserInAuth = (api) => (email) => {
    const emails = api.knownUsers().map((i) => i.email);
    return emails.includes(email);
};
const emailIsValidFormat = (_api) => (email) => {
    return validate(email);
};
const emailHasCorrectPassword = (api) => (email, password) => {
    const config = api.knownUsers().find((i) => i.email === email);
    return config ? config.password === password : false;
};
const userUid = (api) => (email) => {
    const user = api.findKnownUser('email', email);
    return user ? user.uid : uuid();
};
const emailValidationAllowed = (api) => () => {
    return api.hasProvider('emailPassword');
};

const claims = (api) => ({
    /**
     * Sets additional developer claims on an existing user identified by the provided uid,
     * typically used to define user roles and levels of access.
     */
    // eslint-disable-next-line @typescript-eslint/require-await
    async setCustomUserClaims(uid, customClaims) {
        api.updateUser(uid, { customClaims });
    },
});

const tokens = (api) => ({
    /**
     * Verifies a Firebase ID token (JWT). If the token is valid, the promise is fulfilled
     * with the token's decoded claims; otherwise, the promise is rejected. An optional
     * flag can be passed to additionally check whether the ID token was revoked.
     *
     * @param idToken The ID token to verify
     * @param checkRevoked Whether to check if the ID token was revoked. This requires an
     * extra request to the Firebase Auth backend to check the tokensValidAfterTime time
     * for the corresponding user. When not specified, this additional check is not applied.
     */
    verifyIdToken(idToken, checkRevoked) {
        throw new Error('not implemented');
    },
});

const users = (api) => ({
    // https://firebase.google.com/docs/auth/admin/manage-users#create_a_user
    // eslint-disable-next-line @typescript-eslint/require-await
    async createUser(properties) {
        const UserRecord = {
            ...properties,
            metadata: {
                lastSignInTime: null,
                creationTime: String(new Date()),
                toJSON() {
                    return JSON.stringify(properties);
                },
            },
            multiFactor: null,
            toJSON: () => null,
            providerData: null,
            emailVerified: false,
        };
        api.addToUserPool(UserRecord);
        return UserRecord;
    },
    /** Updates an existing user (admin-sdk). */
    async updateUser(user, properties) {
        let uid;
        if (typeof user !== 'string' && (isUser(user) || isMockUserRecord(user))) {
            uid = user.uid;
        }
        else {
            uid = user;
        }
        api.updateUser(uid, properties);
        return Promise.resolve(toUserRecord(api.getCurrentUser()));
    },
    async deleteUser(uid) {
        await api.networkDelay();
        api.removeFromUserPool(uid);
    },
    async getUserByEmail(email) {
        await api.networkDelay();
        return toUserRecord(api.findKnownUser('email', email));
    },
    async getUserByPhoneNumber(phoneNumber) {
        return Promise.resolve(toUserRecord(api.findKnownUser('phoneNumber', phoneNumber)));
    },
    async listUsers(maxResults, pageToken) {
        await api.networkDelay();
        return {
            users: maxResults
                ? api.knownUsers().slice(0, maxResults)
                : api.knownUsers(),
        };
    },
});

const implemented = (api) => ({
    ...users(api),
    ...claims(api),
    ...tokens(),
});

const notImplemented = {
//
};

/**
 * Passing the in the management API returns the Firebase Client Auth API
 */
const createAdminAuth = (api) => {
    return {
        ...implemented(api),
        ...notImplemented
    };
};

/**
 * A factory function which generates the appropriate Auth API (for the
 * passed in SDK) as well as a `authManager` API to manipulate the auth state
 * with super-powers.
 */
function createAuth(sdk, mockAuth) {
    const authManager = createAuthManager(sdk);
    authManager.initializeAuth(mockAuth);
    return (isAdminSdk(sdk)
        ? [createAdminAuth(authManager), authManager]
        : [createClientAuth(authManager), authManager]);
}

// TODO: Removed because this state is going to hanlded in `createStore.ts`
// let _listeners: IListener[] = [];
/**
 * **addListener**
 *
 * Adds a listener for watched events; setup by
 * the `query.on()` API call.
 *
 * This listener is
 * pushed onto a private stack but can be interogated
 * with a call to `getListeners()` or if you're only
 * interested in the _paths_ which are being watched
 * you can call `listenerPaths()`.
 */
// TODO: the code below has been moved into `createStore.ts`
// export async function addListener(
//   pathOrQuery: string | SerializedRealTimeQuery<any>,
//   eventType: IRtdbDbEvent,
//   callback: IFirebaseEventHandler,
//   cancelCallbackOrContext?: (err?: Error) => void,
//   context?: IDictionary
// ): Promise<IRtdbDataSnapshot> {
//   const query = (typeof pathOrQuery === 'string'
//     ? new SerializedRealTimeQuery(join(pathOrQuery))
//     : pathOrQuery) as SerializedRealTimeQuery;
//   pathOrQuery = (typeof pathOrQuery === 'string'
//     ? join(pathOrQuery)
//     : query.path) as string;
//   _listeners.push({
//     id: Math.random().toString(36).substr(2, 10),
//     query,
//     eventType,
//     callback,
//     cancelCallbackOrContext,
//     context,
//   });
//   function ref(dbPath: string) {
//     return new Reference(dbPath);
//   }
//   const snapshot = await query
//     .deserialize({ ref })
//     .once(eventType === 'value' ? 'value' : 'child_added');
//   if (eventType === 'value') {
//     callback(snapshot);
//   } else {
//     const list = hashToArray(snapshot.val());
//     if (eventType === 'child_added') {
//       list.forEach((i: IDictionary) =>
//         callback(new SnapShot(join(query.path, i.id), i))
//       );
//     }
//   }
//   return snapshot;
// }
/**
 * **removeListener**
 *
 * Removes an active listener (or multiple if the info provided matches more
 * than one).
 *
 * If you provide the `context` property it will use this to identify
 * the listener, if not then it will use `eventType` (if available) as
 * well as `callback` (if available) to identify the callback(s)
 */
// TODO: This has been moved into  `createStore.ts`
// export function removeListener(
//   eventType?: IRtdbDbEvent,
//   callback?: (snap: IRtdbDataSnapshot, key?: string) => void,
//   context?: IDictionary
// ): number {
//   if (!eventType) {
//     return removeAllListeners();
//   }
//   if (!callback) {
//     const removed = _listeners.filter((l) => l.eventType === eventType);
//     _listeners = _listeners.filter((l) => l.eventType !== eventType);
//     return cancelCallback(removed);
//   }
//   if (!context) {
//     // use eventType and callback to identify
//     const removed = _listeners
//       .filter((l) => l.callback === callback)
//       .filter((l) => l.eventType === eventType);
//     _listeners = _listeners.filter(
//       (l) => l.eventType !== eventType || l.callback !== callback
//     );
//     return cancelCallback(removed);
//   } else {
//     // if we have context then we can ignore other params
//     const removed = _listeners
//       .filter((l) => l.callback === callback)
//       .filter((l) => l.eventType === eventType)
//       .filter((l) => l.context === context);
//     _listeners = _listeners.filter(
//       (l) =>
//         l.context !== context ||
//         l.callback !== callback ||
//         l.eventType !== eventType
//     );
//     return cancelCallback(removed);
//   }
// }
/**
 * internal function responsible for the actual removal of
 * a listener.
 */
// function cancelCallback<TSdk extends ISdk>(
//   removed: IMockListener<TSdk>[]
// ): number {
//   let count = 0;
//   removed.forEach((l) => {
//     if (typeof l.cancelCallbackOrContext === 'function') {
//       (l.cancelCallbackOrContext as () => any)();
//       count++;
//     }
//   });
//   return count;
// }
// TODO: This method should be rewritten or we have to write tests without depending of this method.
/// Because it is only being used in tests
// export function removeAllListeners<TSdk extends ISdk>(store: IMockStore<TSdk>) {
//   const _listeners = store.getAllListeners();
//   return (): number => {
//     const howMany = cancelCallback(_listeners);
//     _listeners = [];
//     return howMany;
//   };
// }
/**
 * **listenerCount**
 *
 * Provides a numberic count of listeners on the database.
 * Optionally you can state the `EventType` and get a count
 * of only this type of event.
 */
// export function listenerCount(type?: IRtdbDbEvent) {
//   return type
//     ? _listeners.filter((l) => l.eventType === type).length
//     : _listeners.length;
// }
/**
 * **listenerPaths**
 *
 * Provides a list of _paths_ in the database which have listeners
 * attached to them. Optionally you can state the `EventType` and filter down to
 * only this type of event or "set of events".
 *
 * You can also just state "child" as the event and it will resolve to all child
 * events: `[ 'child_added', 'child_changed', 'child_removed', 'child_moved' ]`
 */
// eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
function listenerPaths(store) {
    const _listeners = store.getAllListeners();
    return (lookFor) => {
        if (lookFor && !Array.isArray(lookFor)) {
            lookFor =
                lookFor === 'child'
                    ? ['child_added', 'child_changed', 'child_removed', 'child_moved']
                    : [lookFor];
        }
        return lookFor
            ? _listeners
                .filter((l) => lookFor.includes(l.eventType))
                .map((l) => l.query.path)
            : _listeners.map((l) => l.query.path);
    };
}
/**
 * **getListeners**
 *
 * Returns the list of listeners.Optionally you can state the `EventType` and
 * filter down to only this type of event or "set of events".
 *
 * You can also just state "child" as the event and it will resolve to all child
 * events: `[ 'child_added', 'child_changed', 'child_removed', 'child_moved' ]`
 */
function getListeners(store) {
    const _listeners = store.getAllListeners();
    return (...lookFor) => {
        const childEvents = [
            'child_added',
            'child_changed',
            'child_removed',
            'child_moved',
        ];
        const allEvents = childEvents.concat(['value']);
        const events = lookFor.length === 0
            ? allEvents
            : lookFor.length === 1 && lookFor[0] === 'child'
                ? childEvents
                : lookFor;
        return _listeners.filter((l) => events.includes(l.eventType));
    };
}
function keyDidNotPreviouslyExist(e, dbSnapshot) {
    return get(dbSnapshot, e.key) === undefined ? true : false;
}
/**
 * **notify**
 *
 * Based on a dictionary of paths/values it reduces this to events to
 * send to zero or more listeners.
 */
function notify(api) {
    return function (events, dbSnapshot) {
        events.forEach((evt) => {
            const isDeleteEvent = evt.value === null || evt.value === undefined;
            switch (evt.listenerEvent) {
                case 'child_removed':
                    if (isDeleteEvent) {
                        evt.callback(snapshot(api, evt.key, evt.priorValue));
                    }
                    return;
                case 'child_added':
                    if (!isDeleteEvent && keyDidNotPreviouslyExist(evt, dbSnapshot)) {
                        evt.callback(snapshot(api, evt.key, evt.value));
                    }
                    return;
                case 'child_changed':
                    if (!isDeleteEvent) {
                        evt.callback(snapshot(api, evt.key, evt.value));
                    }
                    return;
                case 'child_moved':
                    if (!isDeleteEvent && keyDidNotPreviouslyExist(evt, dbSnapshot)) {
                        // TODO: if we implement sorting then add the previousKey value
                        evt.callback(snapshot(api, evt.key, evt.value));
                    }
                    return;
                case 'value': {
                    const snapKey = snapshot(api, evt.listenerPath, evt.value).key;
                    if (snapKey === evt.key) {
                        // root set
                        evt.callback(snapshot(api, evt.listenerPath, evt.value === null || evt.value === undefined
                            ? undefined
                            : { [evt.key]: evt.value }));
                    }
                    else {
                        // property set
                        const value = evt.value === null ? api.getDb(evt.listenerPath) : evt.value;
                        evt.callback(snapshot(api, evt.listenerPath, value));
                    }
                }
            } // end switch
        });
    };
}
/**
 * **findChildListeners**
 *
 * Finds "child events" listening to a given _parent path_; optionally
 * allowing for specification of the specific `EventType` or `EventType(s)`.
 *
 * @param changePath the _parent path_ that children are detected off of
 * @param eventTypes <optional> the specific child event (or events) to filter down to; if you have more than one then you should be aware that this property is destructured so the calling function should pass in an array of parameters rather than an array as the second parameter
 */
function findChildListeners(store) {
    const _listeners = store.getAllListeners();
    return (changePath, ...eventTypes) => {
        changePath = stripLeadingDot(changePath.replace(/\//g, '.'));
        eventTypes =
            eventTypes.length !== 0
                ? eventTypes
                : ['child_added', 'child_changed', 'child_moved', 'child_removed'];
        const decendants = _listeners
            .filter((l) => eventTypes.includes(l.eventType))
            .filter((l) => changePath.startsWith(dotify(l.query.path)))
            .reduce((acc, listener) => {
            const id = removeDots(changePath
                .replace(listener.query.path, '')
                .split('.')
                .filter((i) => i)[0]);
            const remainingPath = stripLeadingDot(changePath.replace(stripLeadingDot(listener.query.path), ''));
            const changeIsAtRoot = id === remainingPath;
            acc.push({ ...listener, ...{ id, changeIsAtRoot } });
            return acc;
        }, []);
        return decendants;
    };
}
/**
 * Finds all value listeners on a given path or below.
 * Unlike child listeners, Value listeners listen to changes at
 * all points below the registered path.
 *
 * @param path path to root listening point
 */
function findValueListeners(store) {
    const _listeners = store.getAllListeners();
    return (path) => {
        // TODO: Fix this type issue
        return _listeners.filter((l) => join(path).indexOf(join(l.query.path)) !== -1 && l.eventType === 'value');
    };
}

function onDisconnect(_store, _query) {
    const disconnect = {
        cancel: async () => {
            return Promise.resolve();
        },
        remove: async () => {
            return Promise.resolve();
        },
        set: async () => {
            return Promise.resolve();
        },
        setWithPriority: async () => {
            return Promise.resolve();
        },
        update: async () => {
            return Promise.resolve();
        },
    };
    return disconnect;
}

/* eslint-disable @typescript-eslint/no-unsafe-return */
/** an filter function for queries with a `limitToFirst` value */
// eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
function limitToFirst(query) {
    const value = query.identity.limitToFirst;
    return (list) => {
        if (value === undefined) {
            return list;
        }
        return list.slice(0, value);
    };
}
/** an filter function for queries with a `limitToLast` value */
// eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
function limitToLast(query) {
    const value = query.identity.limitToLast;
    return (list) => {
        if (value === undefined) {
            return list;
        }
        return list.slice(-1 * value);
    };
}

function startAt(query) {
    const key = query.identity.startAtKey || query.identity.orderByKey;
    const value = query.identity.startAt;
    return (record) => {
        if (value === undefined) {
            return true;
        }
        return key ? record[key] >= value : record >= value;
    };
}
function endAt(query) {
    const key = query.identity.endAtKey || query.identity.orderByKey;
    const value = query.identity.endAt;
    return (record) => {
        if (value === undefined) {
            return true;
        }
        return key ? record[key] <= value : record <= value;
    };
}
/** a filter function for queries with a `equalTo` value */
function equalTo(query) {
    const key = query.identity.equalToKey || query.identity.orderByKey;
    const value = query.identity.equalTo;
    return (record) => {
        if (value === undefined) {
            return true;
        }
        return key ? record[key] === value : record === value;
    };
}

/* eslint-disable @typescript-eslint/no-explicit-any */
const orderByKey = (list) => {
    const keys = Object.keys(list).sort();
    const hash = {};
    keys.forEach((k) => {
        hash[k] = list[k];
    });
    return hash;
};
const orderByValue = (list, direction = SortOrder.asc) => {
    const values = hashToArray(list).sort((a, b) => a.value > b.value
        ? direction === SortOrder.asc
            ? 1
            : -1
        : direction === SortOrder.asc
            ? -1
            : 1);
    return values.reduce((agg, curr) => {
        agg[curr.id] = curr.value;
        return agg;
    }, {});
};
const sortFn = (query) => 
// eslint-disable-next-line @typescript-eslint/no-unsafe-return
query.identity.orderBy === RtdbOrder.orderByChild
    ? orderByChild(query.identity.orderByKey)
    // eslint-disable-next-line @typescript-eslint/no-unnecessary-type-assertion
    : sortFns[query.identity.orderBy];
function runQuery(query, data) {
    /**
     * A boolean _flag_ to indicate whether the path is of the query points to a Dictionary
     * of Objects. This is indicative of a **Firemodel** list node.
     */
    const isListOfObjects = typeof data === 'object' &&
        Object.keys(data).every((i) => typeof data[i] === 'object');
    const dataIsAScalar = ['string', 'boolean', 'number'].includes(typeof data);
    if (dataIsAScalar) {
        return data;
    }
    const dataIsAnObject = !Array.isArray(data) && typeof data === 'object';
    if (dataIsAnObject && !isListOfObjects) {
        data =
            query.identity.orderBy === 'orderByKey'
                ? orderByKey(data)
                : orderByValue(data);
        // allows non-array data that can come from a 'value' listener
        // to pass through at this point
        const limitToKeys = query.identity.limitToFirst
            ? Object.keys(data).slice(0, query.identity.limitToFirst)
            : query.identity.limitToLast
                ? Object.keys(data).slice(-1 * query.identity.limitToLast)
                : false;
        if (limitToKeys) {
            Object.keys(data).forEach((k) => {
                if (!limitToKeys.includes(k)) {
                    delete data[k];
                }
            });
        }
        return data;
    }
    const dataList = isListOfObjects || dataIsAnObject ? hashToArray(data) : data;
    if (!dataList) {
        return undefined;
    }
    const limitFilter = _limitFilter(query);
    const queryFilter = _queryFilter(query);
    let list;
    console.log(dataList);
    if (Array.isArray(dataList)) {
        list = limitFilter(queryFilter(dataList.sort(sortFn(query))));
    }
    console.log(list);
    // eslint-disable-next-line @typescript-eslint/no-unsafe-return
    return isListOfObjects
        ? // this is list of records to convert back to hash for Firebase compatability
            arrayToHash(list)
        : dataIsAnObject
            ? // if it was an Object but values weren't objects then this is probably
                // a key/value pairing
                list.reduce((agg, curr) => {
                    if (curr.id && curr.value) {
                        // definitely looks like a id/value pairing
                        agg[curr.id] = curr.value;
                    }
                    else if (curr.id) {
                        // has an ID so offset using the id but use the remainder of the hash
                        // as the value
                        const hash = { ...curr };
                        delete hash.id;
                        agg[curr.id] = hash;
                    }
                    else {
                        console.log({
                            message: `Unsure what to do with part of a data structure resulting from the the query: ${JSON.stringify(query.identity)}.\n\nThe item in question was: "${JSON.stringify(curr)}".`,
                            severity: 0,
                        });
                    }
                    return agg;
                }, {})
            : list;
}
function _limitFilter(query) {
    const first = limitToFirst(query);
    const last = limitToLast(query);
    return (list) => {
        return first(last(list));
    };
}
function _queryFilter(query) {
    return (list) => {
        return list
            .filter(equalTo(query))
            .filter(startAt(query))
            .filter(endAt(query));
    };
}

function isMultiPath(data) {
    Object.keys(data).map((d) => {
        if (!d) {
            data[d] = '/';
        }
    });
    const indexesAreStrings = Object.keys(data).every((i) => typeof i === 'string');
    const indexesLookLikeAPath = Object.keys(data).every((i) => i.indexOf('/') !== -1);
    return indexesAreStrings && indexesLookLikeAPath ? true : false;
}
function reference(store, path = '') {
    const serializedQuery = typeof path === "string" ? new SerializedRealTimeQuery(path) : path;
    const query_ = query(store, serializedQuery);
    const ref = {
        endBefore(_value, key) {
            return query_.endBefore(key);
        },
        get() {
            return query_.get();
        },
        startAfter(_value, key) {
            return query_.startAfter(key);
        },
        orderByKey: query_.orderByKey,
        endAt: query_.endAt,
        // eslint-disable-next-line @typescript-eslint/unbound-method
        // endBefore: query_.endBefore,
        equalTo: query_.equalTo,
        // eslint-disable-next-line @typescript-eslint/unbound-method
        // get: query_.get,
        isEqual: query_.isEqual,
        limitToFirst: query_.limitToFirst,
        limitToLast: query_.limitToLast,
        off: query_.off,
        on: query_.on,
        once: query_.once,
        orderByChild: query_.orderByChild,
        // eslint-disable-next-line @typescript-eslint/unbound-method
        orderByPriority: query_.orderByPriority,
        orderByValue: query_.orderByValue,
        get ref() {
            return query_.ref;
        },
        // eslint-disable-next-line @typescript-eslint/unbound-method
        // startAfter: query_.startAfter,
        startAt: query_.startAt,
        get key() {
            return serializedQuery.path
                ? serializedQuery.path.split('/').pop()
                : null;
        },
        child: (path) => {
            return reference(store, path);
        },
        onDisconnect: () => {
            return onDisconnect();
        },
        get parent() {
            return reference(store, !serializedQuery.path || serializedQuery.path.split('/').length === 1
                ? null
                : serializedQuery.path
                    .split('/')
                    .slice(0, serializedQuery.path.split('/').length - 1)
                    .join('/'));
        },
        push: (value, onComplete) => {
            try {
                const key = store.pushDb(serializedQuery.path, value);
                if (onComplete)
                    onComplete(undefined);
                const _ref = reference(store, join(serializedQuery.path, key));
                const _refPromise = Promise.resolve(_ref);
                return { ..._ref, ..._refPromise };
            }
            catch (e) {
                if (onComplete)
                    onComplete(e);
            }
        },
        remove: async (onComplete) => {
            try {
                store.removeDb(serializedQuery.path);
                if (onComplete)
                    onComplete(undefined);
            }
            catch (e) {
                if (onComplete)
                    onComplete(e);
            }
        },
        get root() {
            return reference(store, null);
        },
        set: async (value, onComplete) => {
            try {
                store.setDb(serializedQuery.path, value);
                if (onComplete)
                    onComplete(null);
                return store.networkDelay();
            }
            catch (e) {
                if (onComplete)
                    onComplete(e);
            }
        },
        setPriority: async () => {
            throw new FireMockError(`Setting priorities with setPriority() is not supported in Firemock and in general isn't a good idea with the Real Time Database.`, 'not-supported');
        },
        setWithPriority: async () => {
            throw new FireMockError(`Setting priorities with setWithPriority() is not supported in Firemock and in general isn't a good idea with the Real Time Database.`, 'not-supported');
        },
        toJSON: () => {
            return {
                identity: ref.toString(),
                query: serializedQuery.identity,
            };
        },
        toString: () => {
            return `FireMock::Query@${process.env.FIREBASE_DATA_ROOT_URL}/${JSON.stringify(serializedQuery.identity)}`;
        },
        transaction: async () => {
            return Promise.resolve({
                committed: true,
                snapshot: null,
                toJSON() {
                    return {};
                },
            });
        },
        update: async (values, onComplete) => {
            try {
                if (isMultiPath(values)) {
                    store.multiPathUpdate(values);
                }
                else {
                    store.updateDb(serializedQuery.path, values);
                }
                if (onComplete)
                    onComplete(undefined);
            }
            catch (e) {
                if (onComplete)
                    onComplete(e);
            }
        },
    };
    return ref;
}

function snapshot(store, key, value) {
    const val = () => (Array.isArray(value) ? arrayToHash(value) : value);
    let sortingFunction;
    const snapshotKey = getKey(join(key));
    return {
        get key() {
            return snapshotKey;
        },
        val,
        get ref() {
            return reference(store, snapshotKey);
        },
        toJSON() {
            return JSON.stringify(value);
        },
        child(path) {
            const _value = get(value, path, null);
            return _value ? snapshot(store, path, _value) : null;
        },
        hasChild(path) {
            if (typeof value === 'object') {
                return Object.keys(value).indexOf(path) !== -1;
            }
            return false;
        },
        hasChildren() {
            if (typeof value === 'object') {
                return Object.keys(value).length > 0;
            }
            return false;
        },
        numChildren() {
            if (typeof value === 'object' && value) {
                return Object.keys(value).length;
            }
            return 0;
        },
        exists() {
            return value !== null;
        },
        forEach(actionCb) {
            const cloned = value.slice(0);
            const sorted = cloned.sort(sortingFunction);
            sorted.map((item) => {
                const noId = { ...{}, ...item };
                delete noId.id;
                const halt = actionCb(snapshot(store, item.id, noId));
                if (halt) {
                    return true;
                }
            });
            return false;
        },
        /** NOTE: mocking proxies this call through to val(), no use of "priority" */
        exportVal() {
            return val();
        },
        getPriority() {
            return null;
        },
    };
    // return {
    //   sortingFunction(fn: SortingFunction) {
    //     sortingFunction = fn;
    //   },
    //   ...partialSnapshot,
    // };
}

/* eslint-disable @typescript-eslint/no-explicit-any */
const query = (store, serializedQuery) => {
    const partialQuery = {
        endBefore(_value) {
            return query(store, serializedQuery);
        },
        startAfter(_value, _key) {
            return query(store, serializedQuery);
        },
        toString() {
            return `FireMock::Query@${process.env.FIREBASE_DATA_ROOT_URL}/${JSON.stringify(serializedQuery.identity)}`;
        },
        get: async () => {
            await store.networkDelay();
            const data = store.getDb(serializedQuery.path);
            const results = runQuery(serializedQuery, data);
            // TODO: See how this was implemented before
            return snapshot(store, leafNode(serializedQuery.path), results ? results : null);
        },
        endAt: (value, key) => {
            serializedQuery.endAt(value, key);
            return query(store, serializedQuery);
        },
        equalTo: (value, key) => {
            if (key && serializedQuery.identity.orderBy === RtdbOrder.orderByKey) {
                throw new Error(`You can not use "equalTo(val, key)" with a "key" property defined when using a key sort!`);
            }
            serializedQuery.equalTo(value, key);
            return query(store, serializedQuery);
        },
        isEqual: (other) => {
            return serializedQuery.hashCode() === other._query.hashCode();
        },
        limitToFirst: (limit) => {
            serializedQuery.limitToFirst(limit);
            return query(store, serializedQuery);
        },
        limitToLast: (limit) => {
            serializedQuery.limitToLast(limit);
            return query(store, serializedQuery);
        },
        off: (_eventType, _callback) => {
            console.log('off() not implemented yet on Firemock');
        },
        on: (eventType, callback, cancelCallbackOrContext, context) => {
            store.addListener(serializedQuery, eventType, callback, cancelCallbackOrContext, context);
            return callback;
        },
        once: async (_eventType) => {
            await store.networkDelay();
            const data = store.getDb(serializedQuery.path);
            const results = runQuery(serializedQuery, data);
            console.log(serializedQuery, results);
            // TODO: See how this was implemented before
            return snapshot(store, leafNode(serializedQuery.path), results ? results : null);
        },
        orderByChild: (prop) => {
            serializedQuery.orderByChild(prop);
            return query(store, serializedQuery);
        },
        orderByKey: () => {
            serializedQuery.orderByKey();
            return query(store, serializedQuery);
        },
        orderByPriority: () => {
            // TODO: should we implement? this idea of Priority is code smell these days
            // _query.order();
            return query(store, serializedQuery);
        },
        orderByValue: () => {
            serializedQuery.orderByValue();
            return query(store, serializedQuery);
        },
        get ref() {
            return reference(store, serializedQuery);
        },
        startAt: (value, key) => {
            serializedQuery.startAt(value, key);
            return query(store, serializedQuery);
        },
        toJSON: () => ({
            url: reference(store, serializedQuery).toString(),
            dbConfig: store.config,
        }),
    };
    return partialQuery;
};

/* eslint-disable @typescript-eslint/require-await */
function createFirebaseApp(sdk, store) {
    var _a;
    const name = ((_a = store.config) === null || _a === void 0 ? void 0 : _a.name) || "default";
    const client = {
        name,
        options: {},
        automaticDataCollectionEnabled: false,
        delete: async () => {
            throw new Error("delete() not implemented");
        },
    };
    const admin = {
        name,
        auth() {
            throw new FireMockError(`database.auth() does not have an implementation in Firemock`);
        },
        database() {
            throw new FireMockError(`database.database() does not have an implementation in Firemock`);
        },
        delete() {
            throw new FireMockError(`database.delete() does not have an implementation in Firemock`);
        },
    };
    return (isAdminSdk(sdk) ? admin : client);
}

/* eslint-disable @typescript-eslint/no-empty-function */
function createRtdbAdminMock(store, config) {
    const db = {
        app: createFirebaseApp(SDK.RealTimeAdmin, store),
        ref(path) {
            return reference(store, path);
        },
        refFromURL(url) {
            return reference(store, url);
        },
        // eslint-disable-next-line @typescript-eslint/require-await
        async getRules() {
            return JSON.stringify(store.rules);
        },
        // eslint-disable-next-line @typescript-eslint/require-await
        async getRulesJSON() {
            return store.rules;
        },
        goOffline() { },
        goOnline() { },
        async setRules() { },
    };
    const [auth, authManager] = createAuth(SDK.RealTimeAdmin, config);
    return {
        sdk: SDK.RealTimeAdmin,
        db,
        store,
        auth,
        authManager
    };
}

/* eslint-disable @typescript-eslint/no-unsafe-return */
function createRtdbClientMock(store, config) {
    const db = {
        useEmulator(_host, _port) { },
        get app() {
            return createFirebaseApp(SDK.RealTimeClient, store);
        },
        ref(path) {
            return reference(store, path);
        },
        refFromURL(url) {
            return reference(store, url);
        },
        goOffline() { },
        goOnline() { },
    };
    const [auth, authManager] = createAuth(SDK.RealTimeClient, config);
    return {
        sdk: SDK.RealTimeClient,
        db,
        store,
        auth,
        authManager
    };
}

function getFromDotPath(obj, dotPath) {
    return ['', '/'].includes(dotPath) ? obj : get(obj, dotPath);
}
/**
 * internal function responsible for the actual removal of
 * a listener.
 */
function cancelCallback(removed) {
    let count = 0;
    removed.forEach((l) => {
        if (typeof l.cancelCallbackOrContext === 'function') {
            l.cancelCallbackOrContext();
            count++;
        }
    });
    return count;
}
function createStore(sdk, config, initialState) {
    if (isFirestoreDatabase(sdk)) {
        throw new FireMockError('Currently Firemock is not implemented for the Firestore Database!');
    }
    /**
     * The in-memory dictionary/hash mantained by the mock RTDB to represent
     * the state of the database
     */
    const _state = initialState || {};
    /** flag to indicate whether dispatch events should be fired */
    let _silenceEvents = false;
    /** the artificial time delay used to simulate a real DB's network latency */
    let _networkDelay = NetworkDelay.lazer;
    let _listeners = [];
    const networkDelay$1 = async () => {
        await networkDelay(_networkDelay);
    };
    const setNetworkDelay = (delay) => {
        _networkDelay = delay;
    };
    const addListener = (pathOrQuery, eventType, callback, cancelCallbackOrContext, context) => {
        const query = typeof pathOrQuery === 'string'
            ? // TODO: this needs to be generalized across RTDB and Firestore
                new SerializedRealTimeQuery(pathOrQuery)
            : pathOrQuery;
        const listener = {
            id: Math.random().toString(36).substr(2, 10),
            query,
            eventType,
            callback,
            cancelCallbackOrContext,
            context,
        };
        _listeners.push(listener);
        return listener;
    };
    const groupEventsByWatcher = (data, dbSnapshot) => {
        var _a;
        const _data = dotifyKeys(data);
        const eventPaths = Object.keys(_data).map((i) => dotify(i));
        const response = [];
        const relativePath = (full, partial) => {
            return full.replace(partial, '');
        };
        const justKey = (obj) => (obj ? Object.keys(obj)[0] : null);
        const justValue = (obj) => justKey(obj) ? obj[justKey(obj)] : null;
        (_a = api.getAllListeners()) === null || _a === void 0 ? void 0 : _a.forEach((listener) => {
            const eventPathsUnderListener = eventPaths.filter((path) => path.includes(dotify(listener.query.path)));
            if (eventPathsUnderListener.length === 0) {
                // if there are no listeners then there's nothing to do
                return;
            }
            const paths = [];
            const listenerPath = dotify(listener.query.path);
            const changeObject = eventPathsUnderListener.reduce((changes, path) => {
                paths.push(path);
                if (dotify(listener.query.path) === path) {
                    changes = _data[path];
                }
                else {
                    set(changes, dotify(relativePath(path, listenerPath)), _data[path]);
                }
                return changes;
            }, {});
            const key = listener.eventType === 'value'
                ? changeObject
                    ? justKey(changeObject)
                    : listener.query.path.split('.').pop()
                : dotify(join(slashNotation(listener.query.path), justKey(changeObject)));
            const newResponse = {
                listenerId: listener.id,
                listenerPath,
                listenerEvent: listener.eventType,
                callback: listener.callback,
                eventPaths: paths,
                key,
                changes: justValue(changeObject),
                value: listener.eventType === 'value'
                    ? api.getDb(listener.query.path)
                    : api.getDb(key),
                priorValue: listener.eventType === 'value'
                    ? getFromDotPath(dbSnapshot, listener.query.path)
                    : justValue(getFromDotPath(dbSnapshot, listener.query.path)),
            };
            response.push(newResponse);
        });
        return response;
    };
    const removeAllListeners = () => {
        const howMany = cancelCallback(_listeners);
        _listeners = [];
        return howMany;
    };
    const getAllListeners = () => {
        return _listeners;
    };
    const api = {
        api: isAdminSdk(sdk)
            ? ApiKind.admin
            : ApiKind.client,
        config,
        state: _state,
        networkDelay: networkDelay$1,
        setNetworkDelay,
        addListener,
        removeListener(eventType, callback, context) {
            if (!eventType) {
                return removeAllListeners();
            }
            if (!callback) {
                const removed = _listeners.filter((l) => l.eventType === eventType);
                _listeners = _listeners.filter((l) => l.eventType !== eventType);
                return cancelCallback(removed);
            }
            if (!context) {
                // use eventType and callback to identify
                const removed = _listeners
                    .filter((l) => l.callback === callback)
                    .filter((l) => l.eventType === eventType);
                _listeners = _listeners.filter((l) => l.eventType !== eventType || l.callback !== callback);
                return cancelCallback(removed);
            }
            else {
                // if we have context then we can ignore other params
                const removed = _listeners
                    .filter((l) => l.callback === callback)
                    .filter((l) => l.eventType === eventType)
                    .filter((l) => l.context === context);
                _listeners = _listeners.filter((l) => l.context !== context ||
                    l.callback !== callback ||
                    l.eventType !== eventType);
                return cancelCallback(removed);
            }
        },
        removeAllListeners,
        getAllListeners,
        listenerPaths(lookFor) {
            if (lookFor && !Array.isArray(lookFor)) {
                lookFor =
                    lookFor === 'child'
                        ? ['child_added', 'child_changed', 'child_removed', 'child_moved']
                        : [lookFor];
            }
            return lookFor
                ? _listeners
                    .filter((l) => lookFor.includes(l.eventType))
                    .map((l) => l.query.path)
                : _listeners.map((l) => l.query.path);
        },
        silenceEvents: () => {
            _silenceEvents = true;
        },
        restoreEvents: () => {
            _silenceEvents = false;
        },
        shouldSendEvents: () => {
            return !_silenceEvents;
        },
        clearDb() {
            const keys = Object.keys(_state);
            keys.forEach((key) => delete _state[key]);
        },
        getDb(path) {
            const dotifyPath = path ? dotify(path) : undefined;
            return (dotifyPath ? getFromDotPath(_state, dotifyPath) : _state);
        },
        setDb(path, value, silent = false) {
            const dotPath = join(path);
            const oldRef = getFromDotPath(_state, dotPath);
            const oldValue = typeof oldRef === 'object' ? { ...oldRef, ...{} } : oldRef;
            const isReference = ['object', 'array'].includes(typeof value);
            const dbSnapshot = copy({ ..._state });
            // ignore if no change
            if ((isReference && deepEqual(oldValue, value)) ||
                (!isReference && oldValue === value)) {
                return;
            }
            if (value === null) {
                const parentValue = getFromDotPath(_state, getParent(dotPath));
                if (typeof parentValue === 'object') {
                    delete parentValue[getKey(dotPath)];
                    set(_state, getParent(dotPath), parentValue);
                }
                else {
                    set(_state, dotPath, undefined);
                }
            }
            else {
                set(_state, dotPath, value);
            }
            if (!silent) {
                const events = groupEventsByWatcher({ [dotPath]: value }, dbSnapshot);
                notify(this)(events, dbSnapshot);
            }
        },
        mergeDb(path, value) {
            const dotPath = join(path);
            const old = getFromDotPath(_state, dotPath);
            const merge$1 = merge(old, value);
            this.setDb(path || '', merge$1);
        },
        updateDb(path, value) {
            const dotPath = join(path);
            const oldValue = getFromDotPath(_state, dotPath);
            let changed = true;
            if (typeof value === 'object' &&
                Object.keys(value).every((k) => (oldValue ? oldValue[k] : null) ===
                    value[k])) {
                changed = false;
            }
            if (typeof value !== 'object' && value === oldValue) {
                changed = false;
            }
            if (!changed) {
                return;
            }
            const newValue = typeof oldValue === 'object'
                ? { ...oldValue, ...value }
                : value;
            this.setDb(dotPath, newValue);
        },
        multiPathUpdate(data) {
            const snapshot = copy(_state);
            Object.keys(data).map((key) => {
                const value = data[key];
                const path = key;
                if (getFromDotPath(_state, path) !== value) {
                    // silent set
                    this.setDb(path, value, true);
                }
            });
            const events = groupEventsByWatcher(data, snapshot);
            notify(this)(events, snapshot);
        },
        removeDb(path) {
            if (!this.getDb(path)) {
                return;
            }
            this.setDb(path, null);
        },
        pushDb(path, value) {
            const pushId = key();
            const fullPath = join(path, pushId);
            const valuePlusId = typeof value === 'object' ? { ...value, id: pushId } : value;
            this.setDb(fullPath, valuePlusId);
            return pushId;
        },
        reset() {
            _listeners = [];
            this.clearDb();
        },
    };
    return api;
}

/**
 * A factory object which returns an implementation to handle either RTDB or
 * Firestore database as well as distinguishing between the minor variance between
 * admin and client SDK's
 */
function createDatabase(sdk, 
/**
 * DB and Auth configuration for mock firebase
 */
config = {}, 
/**
 * The initial state of the database
 */
initialState = {}) {
    const store = createStore(sdk, config.db, {
        ...((config === null || config === void 0 ? void 0 : config.db) && isMockConfig(config.db) ? config.db.mockData : {}),
        ...initialState,
    });
    if (isFirestoreDatabase(sdk)) {
        throw new FireMockError(`Attempt to mock a Firestore database failed because this has not been implemented yet!`);
    }
    const db = {
        app: createFirebaseApp(sdk, store),
        ref(path) {
            return reference(store, path);
        },
        refFromURL(url) {
            return reference(store, url);
        },
        // eslint-disable-next-line @typescript-eslint/require-await
        async getRules() {
            return JSON.stringify(store.rules);
        },
        // eslint-disable-next-line @typescript-eslint/require-await
        async getRulesJSON() {
            return store.rules;
        },
        goOffline() {
            throw new Error('goOffline() is not implemented');
        },
        goOnline() {
            throw new Error('goOnline() is not implemented');
        },
        setRules() {
            throw new Error('setRules() is not implemented');
        },
    };
    const [auth, authManager] = createAuth(sdk, config.auth);
    return {
        db,
        auth,
        authManager,
        sdk,
        store,
    };
}

/**
 * A class representing the Firestore Admin SDK's API surface which interacts with the mock
 * database rather than a real DB.
 */
class FirestoreAdminMock {
    constructor(sdk, config, store) {
        this._sdk = sdk;
        this._dbConfig = config;
        this._store = store;
    }
    settings(settings) {
        throw new Error('Method not implemented.');
    }
    collection(collectionPath) {
        throw new Error('Method not implemented.');
    }
    doc(documentPath) {
        throw new Error('Method not implemented.');
    }
    collectionGroup(collectionId) {
        throw new Error('Method not implemented.');
    }
    getAll(...documentRefsOrReadOptions) {
        throw new Error('Method not implemented.');
    }
    recursiveDelete(ref, bulkWriter) {
        throw new Error('Method not implemented.');
    }
    terminate() {
        throw new Error('Method not implemented.');
    }
    listCollections() {
        throw new Error('Method not implemented.');
    }
    runTransaction(updateFunction, transactionOptions) {
        throw new Error('Method not implemented.');
    }
    batch() {
        throw new Error('Method not implemented.');
    }
    bulkWriter(options) {
        throw new Error('Method not implemented.');
    }
    bundle(bundleId) {
        throw new Error('Method not implemented.');
    }
    goOffline() {
        console.log(`The mock database [ ${this._sdk} / ${this._dbConfig.databaseURL} ] has gone offline, triggered by call to goOffline()`);
    }
    goOnline() {
        console.log(`The mock database [ ${this._sdk} / ${this._dbConfig.databaseURL} ] has gone online, triggered by call to goOnline()`);
    }
}

/**
 * A class representing the Firestore Client SDK's API surface which interacts with the mock
 * database rather than a real DB.
 */
class FirestoreClientMock {
    constructor(sdk, config, store) {
        this._sdk = sdk;
        this._dbConfig = config;
        this._store = store;
    }
    settings() {
        throw new Error('Method not implemented.');
    }
    collection(_collectionPath) {
        throw new Error('Method not implemented.');
    }
    doc(_documentPath) {
        throw new Error('Method not implemented.');
    }
    collectionGroup(_collectionId) {
        throw new Error('Method not implemented.');
    }
    getAll(...documentRefsOrReadOptions) {
        throw new Error('Method not implemented.');
    }
    recursiveDelete(ref, bulkWriter) {
        throw new Error('Method not implemented.');
    }
    terminate() {
        throw new Error('Method not implemented.');
    }
    listCollections() {
        throw new Error('Method not implemented.');
    }
    runTransaction(updateFunction, transactionOptions) {
        throw new Error('Method not implemented.');
    }
    batch() {
        throw new Error('Method not implemented.');
    }
    bulkWriter(options) {
        throw new Error('Method not implemented.');
    }
    bundle(bundleId) {
        throw new Error('Method not implemented.');
    }
    goOffline() {
        console.log(`The mock database [ ${this._sdk} / ${this._dbConfig.databaseURL} ] has gone offline, triggered by call to goOffline()`);
    }
    goOnline() {
        console.log(`The mock database [ ${this._sdk} / ${this._dbConfig.databaseURL} ] has gone online, triggered by call to goOnline()`);
    }
    refFromURL(path) {
        // TODO: look into how best to implement this
        throw createError('not-implemented', `refFromURL() is not yet supported on Admin API for RTDB.`);
    }
}

export { FireMockError, FirestoreAdminMock, FirestoreClientMock, SortOrder, completeUserCredential, createAnonymousUser, createAuth, createAuthManager, createDatabase, createFirebaseApp, createRtdbAdminMock, createRtdbClientMock, createStore, fakeApplicationVerifier, findChildListeners, findValueListeners, getListeners, isMockUserRecord, isUser, isUserCredential, isUserRecord, listenerPaths, notify, onDisconnect, query, reference, snapshot, toUser, toUserCredential, toUserRecord };
